from squall import Squall
from squall.router import Router
from squall.testclient import TestClient

app = Squall()

router = Router(prefix="/{segment}")


@router.get("/users/{id}")
def read_user(segment: str, id: str):
    return {"segment": segment, "id": id}


app.include_router(router)


client = TestClient(app)


def test_get():
    response = client.get("/seg/users/foo")
    assert response.status_code == 200, response.text
    assert response.json() == {"segment": "seg", "id": "foo"}



from squall import Request, Squall
from squall.testclient import TestClient

app = Squall(openapi_prefix="/api/v1")


@app.get("/app")
def read_main(request: Request):
    return {"message": "Hello World", "root_path": request.scope.get("root_path")}


client = TestClient(app)

openapi_schema = {
    "openapi": "3.0.2",
    "info": {"title": "Squall", "version": "0.1.0"},
    "paths": {
        "/app": {
            "get": {
                "summary": "Read Main",
                "operationId": "read_main_app_get",
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    }
                },
            }
        }
    },
    "servers": [{"url": "/api/v1"}],
}


def test_openapi():
    response = client.get("/openapi.json")
    assert response.status_code == 200
    assert response.json() == openapi_schema


def test_main():
    response = client.get("/app")
    assert response.status_code == 200
    assert response.json() == {"message": "Hello World", "root_path": "/api/v1"}



import pytest
from squall.router import RootRouter


@pytest.mark.parametrize(
    "path,expected",
    [
        ("/some/item/static", ["some", "item", "static"]),
        (
            "/some/prefix(?P<some>[^/]+)/item/(?P<item>[^/]+)/(?P<more>[^/]+)-suffix",
            ["some", "*", "item", "*", "*"],
        ),
        (
            "/some/prefix{some_id}/item/{item_id}/extra/{extra}-suffix",
            ["some", "*", "item", "*", "extra", "*"],
        ),
        (
            "/some/{some_id}/item/{item_id}/extra/(?P<username>[^/]+)-dw/company/",
            ["some", "*", "item", "*", "extra", "*", "company"],
        ),
    ],
)
def test_get_fast_path_octets(path, expected):
    assert RootRouter._get_fast_path_octets(path) == expected


def test_fast_path_routing(mocker):
    router = RootRouter()
    endpoint1 = mocker.Mock()
    endpoint2 = mocker.Mock()
    endpoint_ws = mocker.Mock()

    router.add_api_route(
        "/some/prefix(?P<some>[^/]+)/item/(?P<item>[^/]+)/(?P<more>[^/]+)-suffix",
        endpoint1,
        methods=["GET", "POST"],
    )
    router.add_api_route(
        "/some/prefix(?P<some>[^/]+)/item/(?P<item>[^/]+)/(?P<more>[^/]+)-suffix",
        endpoint2,
        methods=["PUT", "DELETE"],
    )
    router.add_ws_route(
        "/some/prefix(?P<some>[^/]+)/item/(?P<item>[^/]+)/(?P<more>[^/]+)-suffix",
        endpoint_ws,
    )

    assert (
        router._fast_path_route_http["some"]["*"]["item"]["*"]["*"]["#routes#"]["GET"][
            0
        ].endpoint
        == endpoint1
    )
    assert (
        router._fast_path_route_http["some"]["*"]["item"]["*"]["*"]["#routes#"]["POST"][
            0
        ].endpoint
        == endpoint1
    )
    assert (
        router._fast_path_route_http["some"]["*"]["item"]["*"]["*"]["#routes#"]["PUT"][
            0
        ].endpoint
        == endpoint2
    )
    assert (
        router._fast_path_route_http["some"]["*"]["item"]["*"]["*"]["#routes#"][
            "DELETE"
        ][0].endpoint
        == endpoint2
    )
    assert (
        router._fast_path_route_ws["some"]["*"]["item"]["*"]["*"]["#routes#"][
            0
        ].endpoint
        == endpoint_ws
    )

    assert (
        router.get_http_routes("GET", "/some/prefix1/item/1/1-suffix")[0][0].endpoint
        == endpoint1
    )
    assert (
        router.get_http_routes("POST", "/some/prefix1/item/1/1-suffix")[0][0].endpoint
        == endpoint1
    )
    assert (
        router.get_http_routes("PUT", "/some/prefix1/item/1/1-suffix")[0][0].endpoint
        == endpoint2
    )
    assert (
        router.get_http_routes("DELETE", "/some/prefix1/item/1/1-suffix")[0][0].endpoint
        == endpoint2
    )
    assert (
        router.get_ws_routes("/some/prefix1/item/1/1-suffix")[0].endpoint == endpoint_ws
    )



from dataclasses import dataclass
from typing import List

from squall import Squall
from squall.testclient import TestClient


@dataclass
class Model:
    name: str


app = Squall()


@app.get("/valid1", responses={"500": {"model": int}})
def valid1():
    pass


@app.get("/valid2", responses={"500": {"model": List[int]}})
def valid2():
    pass


@app.get("/valid3", responses={"500": {"model": Model}})
def valid3():
    pass


@app.get("/valid4", responses={"500": {"model": List[Model]}})
def valid4():
    pass


openapi_schema = {
    "openapi": "3.0.2",
    "info": {"title": "Squall", "version": "0.1.0"},
    "components": {
        "schemas": {
            "Model": {
                "type": "object",
                "properties": {"name": {"type": "string"}},
                "required": ["name"],
                "additionalProperties": False,
            },
            "HTTPBadRequestError": {
                "title": "HTTPBadRequestError",
                "type": "object",
                "properties": {
                    "details": {
                        "title": "Detail",
                        "type": "array",
                        "items": {"$ref": "#/components/schemas/ValidationError"},
                    }
                },
            },
        }
    },
    "paths": {
        "/valid1": {
            "get": {
                "summary": "Valid1",
                "operationId": "valid1_valid1_get",
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "500": {
                        "content": {"application/json": {"schema": {"type": "integer"}}}
                    },
                },
            }
        },
        "/valid2": {
            "get": {
                "summary": "Valid2",
                "operationId": "valid2_valid2_get",
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "500": {
                        "content": {
                            "application/json": {
                                "schema": {
                                    "type": "array",
                                    "items": {"type": "integer"},
                                }
                            }
                        }
                    },
                },
            }
        },
        "/valid3": {
            "get": {
                "summary": "Valid3",
                "operationId": "valid3_valid3_get",
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "500": {
                        "content": {
                            "application/json": {
                                "schema": {"$ref": "#/components/schemas/Model"}
                            }
                        }
                    },
                },
            }
        },
        "/valid4": {
            "get": {
                "summary": "Valid4",
                "operationId": "valid4_valid4_get",
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "500": {
                        "content": {
                            "application/json": {
                                "schema": {
                                    "type": "array",
                                    "items": {"$ref": "#/components/schemas/Model"},
                                }
                            }
                        }
                    },
                },
            }
        },
    },
}

client = TestClient(app)


def test_openapi_schema():
    response = client.get("/openapi.json")
    assert response.status_code == 200, response.text
    assert response.json() == openapi_schema


def test_path_operations():
    response = client.get("/valid1")
    assert response.status_code == 200, response.text
    response = client.get("/valid2")
    assert response.status_code == 200, response.text
    response = client.get("/valid3")
    assert response.status_code == 200, response.text
    response = client.get("/valid4")
    assert response.status_code == 200, response.text



from squall import Squall
from squall.testclient import TestClient

swagger_ui_oauth2_redirect_url = "/docs/redirect"

app = Squall(swagger_ui_oauth2_redirect_url=swagger_ui_oauth2_redirect_url)


@app.get("/items/")
async def read_items():
    return {"id": "foo"}


client = TestClient(app)


def test_swagger_ui():
    response = client.get("/docs")
    assert response.status_code == 200, response.text
    assert response.headers["content-type"] == "text/html; charset=utf-8"
    assert "swagger-ui-dist" in response.text
    print(client.base_url)
    assert (
        f"oauth2RedirectUrl: window.location.origin + '{swagger_ui_oauth2_redirect_url}'"
        in response.text
    )


def test_swagger_ui_oauth2_redirect():
    response = client.get(swagger_ui_oauth2_redirect_url)
    assert response.status_code == 200, response.text
    assert response.headers["content-type"] == "text/html; charset=utf-8"
    assert "window.opener.swaggerUIRedirectOauth2" in response.text


def test_response():
    response = client.get("/items/")
    assert response.json() == {"id": "foo"}



from squall.router import RootRouter, Router


def test_sub_router():
    v1_router = Router(prefix="/v1")
    user_router = Router(prefix="/user")
    user_router.add_api_route(
        "/some/prefix(?P<some>[^/]+)/item/(?P<item>[^/]+)/(?P<more>[^/]+)-suffix",
        methods=["GET", "POST"],
        endpoint=lambda *a: {},
    )
    v1_router.include_router(user_router)

    router = RootRouter(prefix="/api")
    router.include_router(v1_router)

    assert router.routes[0].path == "".join(
        [
            "/api",
            "/v1",
            "/user",
            "/some/prefix(?P<some>[^/]+)/item/(?P<item>[^/]+)/(?P<more>[^/]+)-suffix",
        ]
    )



import pytest
from squall import Router, Squall
from squall.routing import APIRoute
from squall.testclient import TestClient

app = Squall()


class APIRouteA(APIRoute):
    x_type = "A"


class APIRouteB(APIRoute):
    x_type = "B"


class APIRouteC(APIRoute):
    x_type = "C"


router_a = Router(route_class=APIRouteA, prefix="/a")
router_b = Router(route_class=APIRouteB, prefix="/b")
router_c = Router(route_class=APIRouteC, prefix="/c")


@router_a.get("/")
def get_a():
    return {"msg": "A"}


@router_b.get("/")
def get_b():
    return {"msg": "B"}


@router_c.get("/")
def get_c():
    return {"msg": "C"}


router_b.include_router(router=router_c)
router_a.include_router(router=router_b)
app.include_router(router=router_a)


client = TestClient(app)

openapi_schema = {
    "openapi": "3.0.2",
    "info": {"title": "Squall", "version": "0.1.0"},
    "paths": {
        "/a/": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    }
                },
                "summary": "Get A",
                "operationId": "get_a_a__get",
            }
        },
        "/a/b/": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    }
                },
                "summary": "Get B",
                "operationId": "get_b_a_b__get",
            }
        },
        "/a/b/c/": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    }
                },
                "summary": "Get C",
                "operationId": "get_c_a_b_c__get",
            }
        },
    },
}


@pytest.mark.parametrize(
    "path,expected_status,expected_response",
    [
        ("/a", 200, {"msg": "A"}),
        ("/a/b", 200, {"msg": "B"}),
        ("/a/b/c", 200, {"msg": "C"}),
        ("/openapi.json", 200, openapi_schema),
    ],
)
def test_get_path(path, expected_status, expected_response):
    response = client.get(path)
    assert response.status_code == expected_status
    assert response.json() == expected_response


def test_route_classes():
    routes = {}
    for r in app.router.routes:
        assert isinstance(r, APIRoute)
        routes[r.path] = r
    assert getattr(routes["/a/"], "x_type") == "A"
    assert getattr(routes["/a/b/"], "x_type") == "B"
    assert getattr(routes["/a/b/c/"], "x_type") == "C"



"""Squall framework"""

__version__ = "0.2.1"

from .applications import Squall as Squall
from .datastructures import UploadFile as UploadFile
from .exceptions import HTTPException as HTTPException
from .params import Body as Body
from .params import Cookie as Cookie
from .params import File as File
from .params import Form as Form
from .params import Header as Header
from .params import Path as Path
from .params import Query as Query
from .requests import Request as Request
from .responses import Response as Response
from .router import Router as Router
from .websockets import WebSocket as WebSocket
from .websockets import WebSocketDisconnect as WebSocketDisconnect



import http
from typing import Optional

import pytest
from squall import Query, Squall
from squall.testclient import TestClient

app = Squall()


@app.get("/query")
def get_query(query=Query(...)):
    return f"foo bar {query}"


@app.get("/query/optional")
def get_query_optional(query=Query(None)):
    if query is None:
        return "foo bar"
    return f"foo bar {query}"


@app.get("/query/int")
def get_query_type(query: int = Query(...)):
    return f"foo bar {query}"


@app.get("/query/int/optional")
def get_query_type_optional(query: Optional[int] = Query(None)):
    if query is None:
        return "foo bar"
    return f"foo bar {query}"


@app.get("/query/int/default")
def get_query_type_int_default(query: int = Query(10)):
    return f"foo bar {query}"


@app.get("/query/param")
def get_query_param(query=Query(None)):
    if query is None:
        return "foo bar"
    return f"foo bar {query}"


@app.get("/query/param-required")
def get_query_param_required(query=Query(...)):
    return f"foo bar {query}"


@app.get("/query/param-required/int")
def get_query_param_required_type(query: int = Query(...)):
    return f"foo bar {query}"


@app.get("/enum-status-code", status_code=http.HTTPStatus.CREATED)
def get_enum_status_code():
    return "foo bar"


client = TestClient(app)

response_missing = {
    "details": [{"loc": ["query_params", "query"], "msg": "Mandatory field missed"}]
}

response_not_valid_int = {
    "details": [
        {"loc": ["query_params", "query"], "msg": "Cast of `int` failed", "val": "foo"}
    ]
}


@pytest.mark.parametrize(
    "path,expected_status,expected_response",
    [
        ("/query", 400, response_missing),
        ("/query?query=baz", 200, "foo bar baz"),
        ("/query?not_declared=baz", 400, response_missing),
        ("/query/optional", 200, "foo bar"),
        ("/query/optional?query=baz", 200, "foo bar baz"),
        ("/query/optional?not_declared=baz", 200, "foo bar"),
        ("/query/int", 400, response_missing),
        ("/query/int?query=42", 200, "foo bar 42"),
        ("/query/int?query=foo", 400, response_not_valid_int),
        ("/query/int?not_declared=baz", 400, response_missing),
        ("/query/int/optional", 200, "foo bar"),
        ("/query/int/optional?query=50", 200, "foo bar 50"),
        ("/query/int/optional?query=foo", 400, response_not_valid_int),
        ("/query/int/default", 200, "foo bar 10"),
        ("/query/int/default?query=50", 200, "foo bar 50"),
        ("/query/int/default?query=foo", 400, response_not_valid_int),
    ],
)
def test_get_path(path, expected_status, expected_response):
    response = client.get(path)
    assert response.status_code == expected_status
    assert response.json() == expected_response



from squall import Router, Squall
from squall.testclient import TestClient

app = Squall()
router = Router()


@router.get("/a", responses={501: {"description": "Error 1"}})
async def a():
    return "a"


@router.get(
    "/b",
    responses={
        502: {"description": "Error 2"},
        "4XX": {"description": "Error with range, upper"},
    },
)
async def b():
    return "b"


@router.get(
    "/c",
    responses={
        "400": {"description": "Error with str"},
        "5XX": {"description": "Error with range"},
        "default": {"description": "A default response"},
    },
)
async def c():
    return "c"


app.include_router(router)

openapi_schema = {
    "openapi": "3.0.2",
    "info": {"title": "Squall", "version": "0.1.0"},
    "paths": {
        "/a": {
            "get": {
                "responses": {
                    "501": {"description": "Error 1"},
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                },
                "summary": "A",
                "operationId": "a_a_get",
            }
        },
        "/b": {
            "get": {
                "responses": {
                    "502": {"description": "Error 2"},
                    "4XX": {"description": "Error with range, upper"},
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                },
                "summary": "B",
                "operationId": "b_b_get",
            }
        },
        "/c": {
            "get": {
                "responses": {
                    "400": {"description": "Error with str"},
                    "5XX": {"description": "Error with range"},
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "default": {"description": "A default response"},
                },
                "summary": "C",
                "operationId": "c_c_get",
            }
        },
    },
}

client = TestClient(app)


def test_openapi_schema():
    response = client.get("/openapi.json")
    assert response.status_code == 200, response.text
    assert response.json() == openapi_schema


def test_a():
    response = client.get("/a")
    assert response.status_code == 200, response.text
    assert response.json() == "a"


def test_b():
    response = client.get("/b")
    assert response.status_code == 200, response.text
    assert response.json() == "b"


def test_c():
    response = client.get("/c")
    assert response.status_code == 200, response.text
    assert response.json() == "c"



from dataclasses import dataclass, field
from typing import List, Optional

import pytest
from squall import Squall
from squall.exceptions import ResponsePayloadValidationError
from squall.testclient import TestClient

app = Squall()


@dataclass
class Item:
    name: str = field()
    price: Optional[float] = None
    owner_ids: Optional[List[int]] = None


@app.get("/items/valid", response_model=Item)
def get_valid():
    return {"name": "valid", "price": 1.0}


@app.get("/items/wrong_type", response_model=Item)
def get_coerce():
    return {"name": "coerce", "price": "1.0"}


@app.get("/items/validlist", response_model=List[Item])
def get_validlist():
    return [
        {"name": "foo"},
        {"name": "bar", "price": 1.0},
        {"name": "baz", "price": 2.0, "owner_ids": [1, 2, 3]},
    ]


client = TestClient(app)


def test_valid():
    response = client.get("/items/valid")
    response.raise_for_status()
    assert response.json() == {"name": "valid", "price": 1.0, "owner_ids": None}


def test_coerce():
    with pytest.raises(ResponsePayloadValidationError):
        client.get("/items/wrong_type")


def test_validlist():
    response = client.get("/items/validlist")
    response.raise_for_status()
    assert response.json() == [
        {"name": "foo", "price": None, "owner_ids": None},
        {"name": "bar", "price": 1.0, "owner_ids": None},
        {"name": "baz", "price": 2.0, "owner_ids": [1, 2, 3]},
    ]



from dataclasses import dataclass
from typing import List, Optional

from apischema import validator
from squall import Squall
from squall.testclient import TestClient

app = Squall()


@dataclass
class Item:
    age: float
    name: Optional[str] = None

    @validator
    def age_positive(self):
        if self.age <= 0:
            raise ValueError("Age should be greater than 0")


@app.post("/items/")
def save_item_no_body(item: List[Item]):
    return {"item": item}


client = TestClient(app)


openapi_schema = {
    "openapi": "3.0.2",
    "info": {"title": "Squall", "version": "0.1.0"},
    "components": {
        "schemas": {
            "Item": {
                "type": "object",
                "properties": {
                    "age": {"type": "number"},
                    "name": {"type": ["string", "null"], "default": None},
                },
                "required": ["age"],
                "additionalProperties": False,
            },
            "ValidationError": {
                "title": "ValidationError",
                "type": "object",
                "properties": {
                    "loc": {
                        "title": "Location",
                        "type": "array",
                        "items": {"type": "string"},
                    },
                    "msg": {"title": "Message", "type": "string"},
                    "type": {"title": "Error Type", "type": "string"},
                },
                "required": ["loc", "msg", "type"],
            },
            "HTTPValidationError": {
                "title": "HTTPValidationError",
                "type": "object",
                "properties": {
                    "detail": {
                        "title": "Detail",
                        "type": "array",
                        "items": {"$ref": "#/components/schemas/ValidationError"},
                    }
                },
            },
            "HTTPBadRequestError": {
                "title": "HTTPBadRequestError",
                "type": "object",
                "properties": {
                    "details": {
                        "title": "Detail",
                        "type": "array",
                        "items": {"$ref": "#/components/schemas/ValidationError"},
                    }
                },
            },
        }
    },
    "paths": {
        "/items/": {
            "post": {
                "summary": "Save Item No Body",
                "operationId": "save_item_no_body_items__post",
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "422": {
                        "description": "Request Body Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "requestBody": {
                    "required": True,
                    "content": {
                        "application/json": {
                            "schema": {
                                "type": "array",
                                "items": {"$ref": "#/components/schemas/Item"},
                            }
                        }
                    },
                },
            }
        }
    },
}

single_error = {"details": [{"loc": [0], "msg": "Age should be greater than 0"}]}

multiple_errors = {
    "details": [
        {"loc": [0, "age"], "msg": "expected type number, found string"},
        {"loc": [1, "age"], "msg": "expected type number, found string"},
    ]
}


def test_openapi_schema():
    response = client.get("/openapi.json")
    assert response.status_code == 200, response.text
    assert response.json() == openapi_schema


def test_put_correct_body():
    response = client.post("/items/", json=[{"name": "Foo", "age": 5}])
    assert response.status_code == 200, response.text
    assert response.json() == {"item": [{"name": "Foo", "age": 5}]}


def test_validation_error():
    response = client.post("/items/", json=[{"name": "Foo", "age": -1.0}])
    assert response.status_code == 422, response.text
    assert response.json() == single_error


def test_put_incorrect_body_multiple():
    response = client.post("/items/", json=[{"age": "five"}, {"age": "six"}])
    assert response.status_code == 422, response.text
    assert response.json() == multiple_errors



from squall import Squall
from squall.testclient import TestClient

from .utils import needs_py39


@needs_py39
def test_typing():
    checks = [
        (list[int], [1, 2, 3]),
        (dict[str, list[int]], {"a": [1, 2, 3], "b": [4, 5, 6]}),
        (tuple[int, ...], [1, 2, 3]),
        (set[int], [1, 2, 3]),  # `set` is converted to `tuple`
    ]

    for test_type, expect in checks:
        app = Squall()

        @app.post("/", response_model=test_type)
        def post_endpoint():
            return expect

        res = TestClient(app).post("/", json=expect)
        assert res.status_code == 200, res.json()
        assert res.json() == expect



import uuid
from dataclasses import dataclass

import pytest
from squall import Squall
from squall.testclient import TestClient

app = Squall()


class MyUuid:
    def __init__(self, uuid_string: str):
        self.uuid = uuid_string

    def __str__(self):
        return self.uuid

    @property  # type: ignore
    def __class__(self):
        return uuid.UUID

    @property
    def __dict__(self):
        """Spoof a missing __dict__ by raising TypeError, this is how
        asyncpg.pgroto.pgproto.UUID behaves"""
        raise TypeError("vars() argument must have __dict__ attribute")


@app.get("/fast_uuid")
def return_fast_uuid():
    # I don't want to import asyncpg for this test so I made my own UUID
    # Import asyncpg and uncomment the two lines below for the actual bug

    # from asyncpg.pgproto import pgproto
    # asyncpg_uuid = pgproto.UUID("a10ff360-3b1e-4984-a26f-d3ab460bdb51")

    asyncpg_uuid = MyUuid("a10ff360-3b1e-4984-a26f-d3ab460bdb51")
    assert isinstance(asyncpg_uuid, uuid.UUID)
    assert type(asyncpg_uuid) != uuid.UUID
    with pytest.raises(TypeError):
        vars(asyncpg_uuid)
    return {"fast_uuid": asyncpg_uuid}


@dataclass
class SomeCustomClass:
    a_uuid: MyUuid


@app.get("/get_custom_class")
def return_some_user():
    return SomeCustomClass(a_uuid=MyUuid("b8799909-f914-42de-91bc-95c819218d01"))


client = TestClient(app)


@pytest.mark.skip(reason="Mechanism for serialisation needed")
def test_dt():
    with client:
        response_simple = client.get("/fast_uuid")
        response_class = client.get("/get_custom_class")

    assert response_simple.json() == {
        "fast_uuid": "a10ff360-3b1e-4984-a26f-d3ab460bdb51"
    }

    assert response_class.json() == {"a_uuid": "b8799909-f914-42de-91bc-95c819218d01"}



import http
from typing import Optional

from squall import Path, Query, Squall
from squall.params import Num, Str

app = Squall()


@app.add_api("/api_route")
def non_operation():
    return {"message": "Hello World"}


def non_decorated_route():
    return {"message": "Hello World"}


app.add_api_route("/non_decorated_route", non_decorated_route)


@app.get("/text")
def get_text():
    return "Hello World"


@app.get("/path/{item_id}")
def get_id(item_id):
    return item_id


@app.get("/path/str/{item_id}")
def get_str_id(item_id: str):
    return item_id


@app.get("/path/int/{item_id}")
def get_int_id(item_id: int):
    return item_id


@app.get("/path/float/{item_id}")
def get_float_id(item_id: float):
    return item_id


@app.get("/path/param/{item_id}")
def get_path_param_id(item_id: Optional[str] = Path(None)):
    return item_id


@app.get("/path/param-required/{item_id}")
def get_path_param_required_id(item_id: str = Path(...)):
    return item_id


@app.get("/path/param-minlength/{item_id}")
def get_path_param_min_len(item_id: str = Path(..., valid=Str(min_len=3))):
    return item_id


@app.get("/path/param-maxlength/{item_id}")
def get_path_param_max_len(item_id: str = Path(..., valid=Str(max_len=3))):
    return item_id


@app.get("/path/param-min_maxlength/{item_id}")
def get_path_param_min_max_len(
    item_id: str = Path(..., valid=Str(max_len=3, min_len=2))
):
    return item_id


@app.get("/path/param-gt/{item_id}")
def get_path_param_gt(item_id: float = Path(..., valid=Num(gt=3))):
    return item_id


@app.get("/path/param-gt0/{item_id}")
def get_path_param_gt0(item_id: float = Path(..., valid=Num(gt=0))):
    return item_id


@app.get("/path/param-ge/{item_id}")
def get_path_param_ge(item_id: float = Path(..., valid=Num(ge=3))):
    return item_id


@app.get("/path/param-lt/{item_id}")
def get_path_param_lt(item_id: float = Path(..., valid=Num(lt=3))):
    return item_id


@app.get("/path/param-lt0/{item_id}")
def get_path_param_lt0(item_id: float = Path(..., valid=Num(lt=0))):
    return item_id


@app.get("/path/param-le/{item_id}")
def get_path_param_le(item_id: float = Path(..., valid=Num(le=3))):
    return item_id


@app.get("/path/param-lt-gt/{item_id}")
def get_path_param_lt_gt(item_id: float = Path(..., valid=Num(lt=3, gt=1))):
    return item_id


@app.get("/path/param-le-ge/{item_id}")
def get_path_param_le_ge(item_id: float = Path(..., valid=Num(le=3, ge=1))):
    return item_id


@app.get("/path/param-lt-int/{item_id}")
def get_path_param_lt_int(item_id: int = Path(..., valid=Num(lt=3))):
    return item_id


@app.get("/path/param-gt-int/{item_id}")
def get_path_param_gt_int(item_id: int = Path(..., valid=Num(gt=3))):
    return item_id


@app.get("/path/param-le-int/{item_id}")
def get_path_param_le_int(item_id: int = Path(..., valid=Num(le=3))):
    return item_id


@app.get("/path/param-ge-int/{item_id}")
def get_path_param_ge_int(item_id: int = Path(..., valid=Num(ge=3))):
    return item_id


@app.get("/path/param-lt-gt-int/{item_id}")
def get_path_param_lt_gt_int(item_id: int = Path(..., valid=Num(lt=3, gt=1))):
    return item_id


@app.get("/path/param-le-ge-int/{item_id}")
def get_path_param_le_ge_int(item_id: int = Path(..., valid=Num(le=3, ge=1))):
    return item_id


@app.get("/query")
def get_query(query):
    return f"foo bar {query}"


@app.get("/query/optional")
def get_query_optional(query=None):
    if query is None:
        return "foo bar"
    return f"foo bar {query}"


@app.get("/query/int")
def get_query_type(query: int):
    return f"foo bar {query}"


@app.get("/query/int/optional")
def get_query_type_optional(query: Optional[int] = None):
    if query is None:
        return "foo bar"
    return f"foo bar {query}"


@app.get("/query/int/default")
def get_query_type_int_default(query: int = 10):
    return f"foo bar {query}"


@app.get("/query/param")
def get_query_param(query=Query(None)):
    if query is None:
        return "foo bar"
    return f"foo bar {query}"


@app.get("/query/param-required")
def get_query_param_required(query=Query(...)):
    return f"foo bar {query}"


@app.get("/query/param-required/int")
def get_query_param_required_type(query: int = Query(...)):
    return f"foo bar {query}"


@app.get("/enum-status-code", status_code=http.HTTPStatus.CREATED)
def get_enum_status_code():
    return "foo bar"



import logging
import subprocess
import sys
from collections import Counter, defaultdict
from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import Container, DefaultDict, Dict, List, Optional, Set

import httpx
import yaml
from github import Github
from pydantic import BaseModel, BaseSettings, SecretStr

github_graphql_url = "https://api.github.com/graphql"

issues_query = """
query Q($after: String) { 
  repository(name: "squall", owner: "stas-dubrovskyi") {
    issues(first: 100, after: $after) {
      edges {
        cursor
        node {
          number
          author {
            login
            avatarUrl
            url
          }
          title
          createdAt
          state
          comments(first: 100) {
            nodes {
              createdAt
              author {
                login
                avatarUrl
                url
              }
            }
          }
        }
      }
    }
  }
}
"""

prs_query = """
query Q($after: String) { 
  repository(name: "squall", owner: "stas-dubrovskyi") {
    pullRequests(first: 100, after: $after) {
      edges {
        cursor
        node {
          number
          labels(first: 100) {
            nodes {
              name
            }
          }
          author {
            login
            avatarUrl
            url
          }
          title
          createdAt
          state
          comments(first: 100) {
            nodes {
              createdAt
              author {
                login
                avatarUrl
                url
              }
            }
          }
          reviews(first:100) {
            nodes {
              author {
                login
                avatarUrl
                url
              }
              state
            }
          }
        }
      }
    }
  }
}
"""

sponsors_query = """
query Q($after: String) {
  user(login: "stas-dubrovskyi") {
    sponsorshipsAsMaintainer(first: 100, after: $after) {
      edges {
        cursor
        node {
          sponsorEntity {
            ... on Organization {
              login
              avatarUrl
              url
            }
            ... on User {
              login
              avatarUrl
              url
            }
          }
          tier {
            name
            monthlyPriceInDollars
          }
        }
      }
    }
  }
}
"""


class Author(BaseModel):
    login: str
    avatarUrl: str
    url: str


class CommentsNode(BaseModel):
    createdAt: datetime
    author: Optional[Author] = None


class Comments(BaseModel):
    nodes: List[CommentsNode]


class IssuesNode(BaseModel):
    number: int
    author: Optional[Author] = None
    title: str
    createdAt: datetime
    state: str
    comments: Comments


class IssuesEdge(BaseModel):
    cursor: str
    node: IssuesNode


class Issues(BaseModel):
    edges: List[IssuesEdge]


class IssuesRepository(BaseModel):
    issues: Issues


class IssuesResponseData(BaseModel):
    repository: IssuesRepository


class IssuesResponse(BaseModel):
    data: IssuesResponseData


class LabelNode(BaseModel):
    name: str


class Labels(BaseModel):
    nodes: List[LabelNode]


class ReviewNode(BaseModel):
    author: Optional[Author] = None
    state: str


class Reviews(BaseModel):
    nodes: List[ReviewNode]


class PullRequestNode(BaseModel):
    number: int
    labels: Labels
    author: Optional[Author] = None
    title: str
    createdAt: datetime
    state: str
    comments: Comments
    reviews: Reviews


class PullRequestEdge(BaseModel):
    cursor: str
    node: PullRequestNode


class PullRequests(BaseModel):
    edges: List[PullRequestEdge]


class PRsRepository(BaseModel):
    pullRequests: PullRequests


class PRsResponseData(BaseModel):
    repository: PRsRepository


class PRsResponse(BaseModel):
    data: PRsResponseData


class SponsorEntity(BaseModel):
    login: str
    avatarUrl: str
    url: str


class Tier(BaseModel):
    name: str
    monthlyPriceInDollars: float


class SponsorshipAsMaintainerNode(BaseModel):
    sponsorEntity: SponsorEntity
    tier: Tier


class SponsorshipAsMaintainerEdge(BaseModel):
    cursor: str
    node: SponsorshipAsMaintainerNode


class SponsorshipAsMaintainer(BaseModel):
    edges: List[SponsorshipAsMaintainerEdge]


class SponsorsUser(BaseModel):
    sponsorshipsAsMaintainer: SponsorshipAsMaintainer


class SponsorsResponseData(BaseModel):
    user: SponsorsUser


class SponsorsResponse(BaseModel):
    data: SponsorsResponseData


class Settings(BaseSettings):
    input_token: SecretStr
    input_standard_token: SecretStr
    github_repository: str


def get_graphql_response(
    *, settings: Settings, query: str, after: Optional[str] = None
):
    headers = {"Authorization": f"token {settings.input_token.get_secret_value()}"}
    variables = {"after": after}
    response = httpx.post(
        github_graphql_url,
        headers=headers,
        json={"query": query, "variables": variables, "operationName": "Q"},
    )
    if not response.status_code == 200:
        logging.error(f"Response was not 200, after: {after}")
        logging.error(response.text)
        raise RuntimeError(response.text)
    data = response.json()
    return data


def get_graphql_issue_edges(*, settings: Settings, after: Optional[str] = None):
    data = get_graphql_response(settings=settings, query=issues_query, after=after)
    graphql_response = IssuesResponse.parse_obj(data)
    return graphql_response.data.repository.issues.edges


def get_graphql_pr_edges(*, settings: Settings, after: Optional[str] = None):
    data = get_graphql_response(settings=settings, query=prs_query, after=after)
    graphql_response = PRsResponse.parse_obj(data)
    return graphql_response.data.repository.pullRequests.edges


def get_graphql_sponsor_edges(*, settings: Settings, after: Optional[str] = None):
    data = get_graphql_response(settings=settings, query=sponsors_query, after=after)
    graphql_response = SponsorsResponse.parse_obj(data)
    return graphql_response.data.user.sponsorshipsAsMaintainer.edges


def get_experts(settings: Settings):
    issue_nodes: List[IssuesNode] = []
    issue_edges = get_graphql_issue_edges(settings=settings)

    while issue_edges:
        for edge in issue_edges:
            issue_nodes.append(edge.node)
        last_edge = issue_edges[-1]
        issue_edges = get_graphql_issue_edges(settings=settings, after=last_edge.cursor)

    commentors = Counter()
    last_month_commentors = Counter()
    authors: Dict[str, Author] = {}

    now = datetime.now(tz=timezone.utc)
    one_month_ago = now - timedelta(days=30)

    for issue in issue_nodes:
        issue_author_name = None
        if issue.author:
            authors[issue.author.login] = issue.author
            issue_author_name = issue.author.login
        issue_commentors = set()
        for comment in issue.comments.nodes:
            if comment.author:
                authors[comment.author.login] = comment.author
                if comment.author.login == issue_author_name:
                    continue
                issue_commentors.add(comment.author.login)
        for author_name in issue_commentors:
            commentors[author_name] += 1
            if issue.createdAt > one_month_ago:
                last_month_commentors[author_name] += 1
    return commentors, last_month_commentors, authors


def get_contributors(settings: Settings):
    pr_nodes: List[PullRequestNode] = []
    pr_edges = get_graphql_pr_edges(settings=settings)

    while pr_edges:
        for edge in pr_edges:
            pr_nodes.append(edge.node)
        last_edge = pr_edges[-1]
        pr_edges = get_graphql_pr_edges(settings=settings, after=last_edge.cursor)

    contributors = Counter()
    commentors = Counter()
    reviewers = Counter()
    authors: Dict[str, Author] = {}

    for pr in pr_nodes:
        author_name = None
        if pr.author:
            authors[pr.author.login] = pr.author
            author_name = pr.author.login
        pr_commentors: Set[str] = set()
        pr_reviewers: Set[str] = set()
        for comment in pr.comments.nodes:
            if comment.author:
                authors[comment.author.login] = comment.author
                if comment.author.login == author_name:
                    continue
                pr_commentors.add(comment.author.login)
        for author_name in pr_commentors:
            commentors[author_name] += 1
        for review in pr.reviews.nodes:
            if review.author:
                authors[review.author.login] = review.author
                pr_reviewers.add(review.author.login)
        for reviewer in pr_reviewers:
            reviewers[reviewer] += 1
        if pr.state == "MERGED" and pr.author:
            contributors[pr.author.login] += 1
    return contributors, commentors, reviewers, authors


def get_individual_sponsors(settings: Settings):
    nodes: List[SponsorshipAsMaintainerNode] = []
    edges = get_graphql_sponsor_edges(settings=settings)

    while edges:
        for edge in edges:
            nodes.append(edge.node)
        last_edge = edges[-1]
        edges = get_graphql_sponsor_edges(settings=settings, after=last_edge.cursor)

    tiers: DefaultDict[float, Dict[str, SponsorEntity]] = defaultdict(dict)
    for node in nodes:
        tiers[node.tier.monthlyPriceInDollars][
            node.sponsorEntity.login
        ] = node.sponsorEntity
    return tiers


def get_top_users(
    *,
    counter: Counter,
    min_count: int,
    authors: Dict[str, Author],
    skip_users: Container[str],
):
    users = []
    for commentor, count in counter.most_common(50):
        if commentor in skip_users:
            continue
        if count >= min_count:
            author = authors[commentor]
            users.append(
                {
                    "login": commentor,
                    "count": count,
                    "avatarUrl": author.avatarUrl,
                    "url": author.url,
                }
            )
    return users


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    settings = Settings()
    logging.info(f"Using config: {settings.json()}")
    g = Github(settings.input_standard_token.get_secret_value())
    repo = g.get_repo(settings.github_repository)
    issue_commentors, issue_last_month_commentors, issue_authors = get_experts(
        settings=settings
    )
    contributors, pr_commentors, reviewers, pr_authors = get_contributors(
        settings=settings
    )
    authors = {**issue_authors, **pr_authors}
    maintainers_logins = {"stas-dubrovskyi"}
    bot_names = {"codecov", "github-actions"}
    maintainers = []
    for login in maintainers_logins:
        user = authors[login]
        maintainers.append(
            {
                "login": login,
                "answers": issue_commentors[login],
                "prs": contributors[login],
                "avatarUrl": user.avatarUrl,
                "url": user.url,
            }
        )

    min_count_expert = 10
    min_count_last_month = 3
    min_count_contributor = 4
    min_count_reviewer = 4
    skip_users = maintainers_logins | bot_names
    experts = get_top_users(
        counter=issue_commentors,
        min_count=min_count_expert,
        authors=authors,
        skip_users=skip_users,
    )
    last_month_active = get_top_users(
        counter=issue_last_month_commentors,
        min_count=min_count_last_month,
        authors=authors,
        skip_users=skip_users,
    )
    top_contributors = get_top_users(
        counter=contributors,
        min_count=min_count_contributor,
        authors=authors,
        skip_users=skip_users,
    )
    top_reviewers = get_top_users(
        counter=reviewers,
        min_count=min_count_reviewer,
        authors=authors,
        skip_users=skip_users,
    )

    tiers = get_individual_sponsors(settings=settings)
    keys = list(tiers.keys())
    keys.sort(reverse=True)
    sponsors = []
    for key in keys:
        sponsor_group = []
        for login, sponsor in tiers[key].items():
            sponsor_group.append(
                {"login": login, "avatarUrl": sponsor.avatarUrl, "url": sponsor.url}
            )
        sponsors.append(sponsor_group)

    people = {
        "maintainers": maintainers,
        "experts": experts,
        "last_month_active": last_month_active,
        "top_contributors": top_contributors,
        "top_reviewers": top_reviewers,
    }
    github_sponsors = {
        "sponsors": sponsors,
    }
    people_path = Path("./docs/en/data/people.yml")
    github_sponsors_path = Path("./docs/en/data/github_sponsors.yml")
    people_old_content = people_path.read_text(encoding="utf-8")
    github_sponsors_old_content = github_sponsors_path.read_text(encoding="utf-8")
    new_people_content = yaml.dump(people, sort_keys=False, width=200, allow_unicode=True)
    new_github_sponsors_content = yaml.dump(github_sponsors, sort_keys=False, width=200, allow_unicode=True)
    if people_old_content == new_people_content and github_sponsors_old_content == new_github_sponsors_content:
        logging.info("The Squall People data hasn't changed, finishing.")
        sys.exit(0)
    people_path.write_text(new_people_content, encoding="utf-8")
    github_sponsors_path.write_text(new_github_sponsors_content, encoding="utf-8")
    logging.info("Setting up GitHub Actions git user")
    subprocess.run(["git", "config", "user.name", "github-actions"], check=True)
    subprocess.run(
        ["git", "config", "user.email", "github-actions@github.com"], check=True
    )
    branch_name = "squall-people"
    logging.info(f"Creating a new branch {branch_name}")
    subprocess.run(["git", "checkout", "-b", branch_name], check=True)
    logging.info("Adding updated file")
    subprocess.run(["git", "add", str(people_path)], check=True)
    logging.info("Committing updated file")
    message = "👥 Update Squall People"
    result = subprocess.run(["git", "commit", "-m", message], check=True)
    logging.info("Pushing branch")
    subprocess.run(["git", "push", "origin", branch_name], check=True)
    logging.info("Creating PR")
    pr = repo.create_pull(title=message, body=message, base="master", head=branch_name)
    logging.info(f"Created PR: {pr.numeric}")
    logging.info("Finished")



import typing
from collections.abc import Mapping
from http import cookies as http_cookies
from typing import Any, Dict, Optional

import anyio
import orjson
from starlette.datastructures import URL, Address, FormData, Headers, QueryParams, State
from starlette.formparsers import FormParser, MultiPartParser
from starlette.types import Message, Receive, Scope, Send

try:
    from multipart.multipart import parse_options_header  # type: ignore
except ImportError:  # pragma: nocover
    parse_options_header = None


json_loads = orjson.loads

SERVER_PUSH_HEADERS_TO_COPY = {
    "accept",
    "accept-encoding",
    "accept-language",
    "cache-control",
    "user-agent",
}


def cookie_parser(cookie_string: str) -> typing.Dict[str, str]:
    """
    This function parses a ``Cookie`` HTTP header into a dict of key/value pairs.

    It attempts to mimic browser cookie parsing behavior: browsers and web servers
    frequently disregard the spec (RFC 6265) when setting and reading cookies,
    so we attempt to suit the common scenarios here.

    This function has been adapted from Django 3.1.0.
    Note: we are explicitly _NOT_ using `SimpleCookie.load` because it is based
    on an outdated spec and will fail on lots of input we want to support
    """
    cookie_dict: typing.Dict[str, str] = {}
    for chunk in cookie_string.split(";"):
        key, val = chunk.split("=", 1) if "=" in chunk else ("", chunk)
        key, val = key.strip(), val.strip()
        if key or val:
            # unquote using Python's algorithm.
            cookie_dict[key] = http_cookies._unquote(val)
    return cookie_dict


class ClientDisconnect(Exception):
    pass


class HTTPConnection(Mapping[str, Any]):
    """
    A base class for incoming HTTP connections, that is used to provide
    any functionality that is common to both `Request` and `WebSocket`.
    """

    def __init__(self, scope: Scope) -> None:
        assert scope["type"] in ("http", "websocket")
        self.scope = scope
        self.path_params = scope.get("path_params", {})

    def __getitem__(self, key: str) -> typing.Any:
        return self.scope[key]

    def __iter__(self) -> typing.Iterator[str]:
        return iter(self.scope)

    def __len__(self) -> int:
        return len(self.scope)

    # Don't use the `abc.Mapping.__eq__` implementation.
    # Connection instances should never be considered equal
    # unless `self is other`.
    __eq__ = object.__eq__
    __hash__ = object.__hash__

    @property
    def app(self) -> typing.Any:
        return self.scope["app"]

    @property
    def url(self) -> URL:
        if not hasattr(self, "_url"):
            self._url = URL(scope=self.scope)
        return self._url

    @property
    def base_url(self) -> URL:
        if not hasattr(self, "_base_url"):
            base_url_scope = dict(self.scope)
            base_url_scope["path"] = "/"
            base_url_scope["query_string"] = b""
            base_url_scope["root_path"] = base_url_scope.get(
                "app_root_path", base_url_scope.get("root_path", "")
            )
            self._base_url = URL(scope=base_url_scope)
        return self._base_url

    @property
    def headers(self) -> Headers:
        if not hasattr(self, "_headers"):
            self._headers = Headers(scope=self.scope)
        return self._headers

    @property
    def query_params(self) -> QueryParams:
        if not hasattr(self, "_query_params"):
            self._query_params = QueryParams(self.scope["query_string"])
        return self._query_params

    # @property
    # def path_params(self) -> dict:
    #     return self.scope.get("path_params", {})

    @property
    def cookies(self) -> typing.Dict[str, str]:
        if not hasattr(self, "_cookies"):
            cookies: typing.Dict[str, str] = {}
            cookie_header = self.headers.get("cookie")

            if cookie_header:
                cookies = cookie_parser(cookie_header)
            self._cookies = cookies
        return self._cookies

    @property
    def client(self) -> Address:
        host, port = self.scope.get("client") or (None, None)
        return Address(host=host, port=port)

    @property
    def session(self) -> Any:
        assert (
            "session" in self.scope
        ), "SessionMiddleware must be installed to access request.session"
        return self.scope["session"]

    @property
    def auth(self) -> typing.Any:
        assert (
            "auth" in self.scope
        ), "AuthenticationMiddleware must be installed to access request.auth"
        return self.scope["auth"]

    @property
    def user(self) -> typing.Any:
        assert (
            "user" in self.scope
        ), "AuthenticationMiddleware must be installed to access request.user"
        return self.scope["user"]

    @property
    def state(self) -> State:
        if not hasattr(self, "_state"):
            # Ensure 'state' has an empty dict if it's not already populated.
            self.scope.setdefault("state", {})
            # Create a state instance with a reference to the dict in which it should
            # store info
            self._state = State(self.scope["state"])
        return self._state


async def empty_receive() -> Message:
    raise RuntimeError("Receive channel has not been made available")


async def empty_send(message: Message) -> None:
    raise RuntimeError("Send channel has not been made available")


class Request(HTTPConnection):
    _json: Dict[str, Any]
    _body: bytes

    def __init__(
        self, scope: Scope, receive: Receive = empty_receive, send: Send = empty_send
    ):
        super().__init__(scope)
        assert scope["type"] == "http"
        self._receive = receive
        self._send = send
        self._stream_consumed = False
        self._is_disconnected = False

    @property
    def method(self) -> str:
        method: str = self.scope["method"]
        return method

    @property
    def receive(self) -> Receive:
        return self._receive

    async def stream(self) -> typing.AsyncGenerator[bytes, None]:
        if hasattr(self, "_body"):
            yield self._body
            yield b""
            return

        if self._stream_consumed:
            raise RuntimeError("Stream consumed")

        self._stream_consumed = True
        message_get = dict.get
        while True:
            message = await self._receive()
            if message["type"] == "http.request":
                if body := message_get(message, "body"):
                    yield body
                if not message_get(message, "more_body"):
                    break
            elif message["type"] == "http.disconnect":
                self._is_disconnected = True
                raise ClientDisconnect()
        yield b""

    async def body(self) -> bytes:
        cached_body: Optional[bytes] = getattr(self, "_body", None)
        if cached_body is not None:
            return cached_body

        message_get = dict.get

        chunks = []
        message = await self._receive()
        if message["type"] == "http.request":
            body: bytes = message_get(message, "body", b"")
            if not message_get(message, "more_body"):
                self._body = body
                return body
            chunks.append(body)
        elif message["type"] == "http.disconnect":
            self._is_disconnected = True
            raise ClientDisconnect()

        while True:
            message = await self._receive()
            if message["type"] == "http.request":
                if body := message_get(message, "body", b""):
                    chunks.append(body)
                if not message_get(message, "more_body"):
                    self._body = _body = b"".join(chunks)
                    return _body
            elif message["type"] == "http.disconnect":
                self._is_disconnected = True
                raise ClientDisconnect()

    async def json(self) -> typing.Any:
        if hasattr(self, "_json"):
            return self._json

        if cached_body := getattr(self, "_body", None) is not None:
            self._json = _json = json_loads(cached_body)  # type: ignore
            return _json

        if self._stream_consumed:
            raise RuntimeError("Stream consumed")

        self._stream_consumed = True

        message_get = dict.get

        chunks = []
        message = await self._receive()
        if message["type"] == "http.request":
            body = message_get(message, "body", b"")
            if not message_get(message, "more_body"):
                self._body = body
                self._json = _json = json_loads(body)
                return _json
            chunks.append(body)
        elif message["type"] == "http.disconnect":
            self._is_disconnected = True
            raise ClientDisconnect()

        while True:
            message = await self._receive()
            if message["type"] == "http.request":
                if body := message_get(message, "body"):
                    chunks.append(body)
                    if not message_get(message, "more_body"):
                        self._body = b"".join(chunks)
                        self._json = _json = json_loads(self._body)
                        return _json
            elif message["type"] == "http.disconnect":
                self._is_disconnected = True
                raise ClientDisconnect()

    async def form(self) -> FormData:
        if not hasattr(self, "_form"):
            assert (
                parse_options_header is not None
            ), "The `python-multipart` library must be installed to use form parsing."
            content_type_header = self.headers.get("Content-Type")
            content_type, options = parse_options_header(content_type_header)
            if content_type == b"multipart/form-data":
                multipart_parser = MultiPartParser(self.headers, self.stream())
                self._form = await multipart_parser.parse()
            elif content_type == b"application/x-www-form-urlencoded":
                form_parser = FormParser(self.headers, self.stream())
                self._form = await form_parser.parse()
            else:
                self._form = FormData()
        return self._form

    async def close(self) -> None:
        form = getattr(self, "_form", None)
        if form is not None:
            await form.close()

    async def is_disconnected(self) -> bool:
        if not self._is_disconnected:
            message: Message

            # If message isn't immediately available, move on
            with anyio.CancelScope() as cs:
                cs.cancel()
                message = await self._receive()

            if message.get("type") == "http.disconnect":
                self._is_disconnected = True

        return self._is_disconnected

    async def send_push_promise(self, path: str) -> None:
        if "http.response.push" in self.scope.get("extensions", {}):
            raw_headers = []
            for name in SERVER_PUSH_HEADERS_TO_COPY:
                for value in self.headers.getlist(name):
                    raw_headers.append(
                        (name.encode("latin-1"), value.encode("latin-1"))
                    )
            await self._send(
                {"type": "http.response.push", "path": path, "headers": raw_headers}
            )



from typing import List

import pytest
from apischema import Unsupported
from squall import Squall


class NonPydanticModel:
    pass


# from apischema.schemas import


def test_invalid_response_model_raises():
    with pytest.raises(Unsupported):
        app = Squall()

        @app.get("/", response_model=NonPydanticModel)
        def read_root():
            pass  # pragma: nocover


def test_invalid_response_model_sub_type_raises():
    with pytest.raises(Unsupported):
        app = Squall()

        @app.get("/", response_model=List[NonPydanticModel])
        def read_root():
            pass  # pragma: nocover


@pytest.mark.skip("Implement additional responses validation")
def test_invalid_response_model_in_responses_raises():
    with pytest.raises(Unsupported):
        app = Squall()

        @app.get("/", responses={"500": {"model": NonPydanticModel}})
        def read_root():
            pass  # pragma: nocover


@pytest.mark.skip("Implement additional responses validation")
def test_invalid_response_model_sub_type_in_responses_raises():
    with pytest.raises(Unsupported):
        app = Squall()

        @app.get("/", responses={"500": {"model": List[NonPydanticModel]}})
        def read_root():
            pass  # pragma: nocover



from typing import Any, Dict, Optional

import orjson
from squall.responses import HTMLResponse


def get_swagger_ui_html(
    *,
    openapi_url: str,
    title: str,
    swagger_js_url: str = "https://cdn.jsdelivr.net/npm/swagger-ui-dist@3/swagger-ui-bundle.js",
    swagger_css_url: str = "https://cdn.jsdelivr.net/npm/swagger-ui-dist@3/swagger-ui.css",
    swagger_favicon_url: str = "https://squall.mtag.dev/img/favicon.png",
    oauth2_redirect_url: Optional[str] = None,
    init_oauth: Optional[Dict[str, Any]] = None,
) -> HTMLResponse:

    html = f"""
    <!DOCTYPE html>
    <html>
    <head>
    <link type="text/css" rel="stylesheet" href="{swagger_css_url}">
    <link rel="shortcut icon" href="{swagger_favicon_url}">
    <title>{title}</title>
    </head>
    <body>
    <div id="swagger-ui">
    </div>
    <script src="{swagger_js_url}"></script>
    <!-- `SwaggerUIBundle` is now available on the page -->
    <script>
    const ui = SwaggerUIBundle({{
        url: '{openapi_url}',
    """

    if oauth2_redirect_url:
        html += f"oauth2RedirectUrl: window.location.origin + '{oauth2_redirect_url}',"

    html += """
        dom_id: '#swagger-ui',
        presets: [
        SwaggerUIBundle.presets.apis,
        SwaggerUIBundle.SwaggerUIStandalonePreset
        ],
        layout: "BaseLayout",
        deepLinking: true,
        showExtensions: true,
        showCommonExtensions: true
    })"""

    if init_oauth:
        html += """
        ui.initOAuth({!r})
        """.format(
            orjson.dumps(init_oauth)
        )

    html += """
    </script>
    </body>
    </html>
    """
    return HTMLResponse(html)


def get_redoc_html(
    *,
    openapi_url: str,
    title: str,
    redoc_js_url: str = "https://cdn.jsdelivr.net/npm/redoc@next/bundles/redoc.standalone.js",
    redoc_favicon_url: str = "https://squall.mtag.dev/img/favicon.png",
    with_google_fonts: bool = True,
) -> HTMLResponse:
    html = f"""
    <!DOCTYPE html>
    <html>
    <head>
    <title>{title}</title>
    <!-- needed for adaptive design -->
    <meta charset="utf-8"/>
    <meta name="viewport" content="width=device-width, initial-scale=1">
    """
    if with_google_fonts:
        html += """
    <link href="https://fonts.googleapis.com/css?family=Montserrat:300,400,700|Roboto:300,400,700" rel="stylesheet">
    """
    html += f"""
    <link rel="shortcut icon" href="{redoc_favicon_url}">
    <!--
    ReDoc doesn't change outer page styles
    -->
    <style>
      body {{
        margin: 0;
        padding: 0;
      }}
    </style>
    </head>
    <body>
    <redoc spec-url="{openapi_url}"></redoc>
    <script src="{redoc_js_url}"> </script>
    </body>
    </html>
    """
    return HTMLResponse(html)


def get_swagger_ui_oauth2_redirect_html() -> HTMLResponse:
    html = """
    <!DOCTYPE html>
    <html lang="en-US">
    <body onload="run()">
    </body>
    </html>
    <script>
        'use strict';
        function run () {
            var oauth2 = window.opener.swaggerUIRedirectOauth2;
            var sentState = oauth2.state;
            var redirectUrl = oauth2.redirectUrl;
            var isValid, qp, arr;

            if (/code|token|error/.test(window.location.hash)) {
                qp = window.location.hash.substring(1);
            } else {
                qp = location.search.substring(1);
            }

            arr = qp.split("&")
            arr.forEach(function (v,i,_arr) { _arr[i] = '"' + v.replace('=', '":"') + '"';})
            qp = qp ? JSON.parse('{' + arr.join() + '}',
                    function (key, value) {
                        return key === "" ? value : decodeURIComponent(value)
                    }
            ) : {}

            isValid = qp.state === sentState

            if ((
            oauth2.auth.schema.get("flow") === "accessCode"||
            oauth2.auth.schema.get("flow") === "authorizationCode"
            ) && !oauth2.auth.code) {
                if (!isValid) {
                    oauth2.errCb({
                        authId: oauth2.auth.name,
                        source: "auth",
                        level: "warning",
                        message: "Authorization may be unsafe, passed state was changed in server Passed state wasn't returned from auth server"
                    });
                }

                if (qp.code) {
                    delete oauth2.state;
                    oauth2.auth.code = qp.code;
                    oauth2.callback({auth: oauth2.auth, redirectUrl: redirectUrl});
                } else {
                    let oauthErrorMsg
                    if (qp.error) {
                        oauthErrorMsg = "["+qp.error+"]: " +
                            (qp.error_description ? qp.error_description+ ". " : "no accessCode received from the server. ") +
                            (qp.error_uri ? "More info: "+qp.error_uri : "");
                    }

                    oauth2.errCb({
                        authId: oauth2.auth.name,
                        source: "auth",
                        level: "error",
                        message: oauthErrorMsg || "[Authorization failed]: no accessCode received from the server"
                    });
                }
            } else {
                oauth2.callback({auth: oauth2.auth, token: qp, isValid: isValid, redirectUrl: redirectUrl});
            }
            window.close();
        }
    </script>
        """
    return HTMLResponse(content=html)



# import pytest
# from squall import File, Form, Squall, UploadFile
# from squall.dependencies.utils import (
#     multipart_incorrect_install_error,
#     multipart_not_installed_error,
# )
#
#
# def test_incorrect_multipart_installed_form(monkeypatch):
#     monkeypatch.delattr("multipart.multipart.parse_options_header", raising=False)
#     with pytest.raises(RuntimeError, match=multipart_incorrect_install_error):
#         app = Squall()
#
#         @app.post("/")
#         async def root(username: str = Form(...)):
#             return username  # pragma: nocover
#
#
# def test_incorrect_multipart_installed_file_upload(monkeypatch):
#     monkeypatch.delattr("multipart.multipart.parse_options_header", raising=False)
#     with pytest.raises(RuntimeError, match=multipart_incorrect_install_error):
#         app = Squall()
#
#         @app.post("/")
#         async def root(f: UploadFile = File(...)):
#             return f  # pragma: nocover
#
#
# def test_incorrect_multipart_installed_file_bytes(monkeypatch):
#     monkeypatch.delattr("multipart.multipart.parse_options_header", raising=False)
#     with pytest.raises(RuntimeError, match=multipart_incorrect_install_error):
#         app = Squall()
#
#         @app.post("/")
#         async def root(f: bytes = File(...)):
#             return f  # pragma: nocover
#
#
# def test_incorrect_multipart_installed_multi_form(monkeypatch):
#     monkeypatch.delattr("multipart.multipart.parse_options_header", raising=False)
#     with pytest.raises(RuntimeError, match=multipart_incorrect_install_error):
#         app = Squall()
#
#         @app.post("/")
#         async def root(username: str = Form(...), password: str = Form(...)):
#             return username  # pragma: nocover
#
#
# def test_incorrect_multipart_installed_form_file(monkeypatch):
#     monkeypatch.delattr("multipart.multipart.parse_options_header", raising=False)
#     with pytest.raises(RuntimeError, match=multipart_incorrect_install_error):
#         app = Squall()
#
#         @app.post("/")
#         async def root(username: str = Form(...), f: UploadFile = File(...)):
#             return username  # pragma: nocover
#
#
# def test_no_multipart_installed(monkeypatch):
#     monkeypatch.delattr("multipart.__version__", raising=False)
#     with pytest.raises(RuntimeError, match=multipart_not_installed_error):
#         app = Squall()
#
#         @app.post("/")
#         async def root(username: str = Form(...)):
#             return username  # pragma: nocover
#
#
# def test_no_multipart_installed_file(monkeypatch):
#     monkeypatch.delattr("multipart.__version__", raising=False)
#     with pytest.raises(RuntimeError, match=multipart_not_installed_error):
#         app = Squall()
#
#         @app.post("/")
#         async def root(f: UploadFile = File(...)):
#             return f  # pragma: nocover
#
#
# def test_no_multipart_installed_file_bytes(monkeypatch):
#     monkeypatch.delattr("multipart.__version__", raising=False)
#     with pytest.raises(RuntimeError, match=multipart_not_installed_error):
#         app = Squall()
#
#         @app.post("/")
#         async def root(f: bytes = File(...)):
#             return f  # pragma: nocover
#
#
# def test_no_multipart_installed_multi_form(monkeypatch):
#     monkeypatch.delattr("multipart.__version__", raising=False)
#     with pytest.raises(RuntimeError, match=multipart_not_installed_error):
#         app = Squall()
#
#         @app.post("/")
#         async def root(username: str = Form(...), password: str = Form(...)):
#             return username  # pragma: nocover
#
#
# def test_no_multipart_installed_form_file(monkeypatch):
#     monkeypatch.delattr("multipart.__version__", raising=False)
#     with pytest.raises(RuntimeError, match=multipart_not_installed_error):
#         app = Squall()
#
#         @app.post("/")
#         async def root(username: str = Form(...), f: UploadFile = File(...)):
#             return username  # pragma: nocover



import typing
from asyncio import iscoroutinefunction
from typing import Any, Callable, Dict, List, Optional, Sequence, Type, Union

from squall import router
from squall.compression import Compression
from squall.concurrency import run_in_threadpool
from squall.datastructures import Default
from squall.errors import get_default_debug_response
from squall.exception_handlers import (
    http_exception_handler,
    request_head_validation_exception_handler,
    request_payload_validation_exception_handler,
)
from squall.exceptions import (
    HTTPException,
    RequestHeadValidationError,
    RequestPayloadValidationError,
)
from squall.lifespan import LifespanContext, lifespan
from squall.logger import logger
from squall.openapi.docs import (
    get_redoc_html,
    get_swagger_ui_html,
    get_swagger_ui_oauth2_redirect_html,
)
from squall.openapi.utils import get_openapi
from squall.requests import Request
from squall.responses import HTMLResponse, JSONResponse, PlainTextResponse, Response
from squall.routing import APIRoute, WebSocketRoute
from squall.types import AnyFunc, ASGIApp, Receive, Scope, Send
from starlette.datastructures import State
from starlette.middleware import Middleware
from starlette.middleware.base import BaseHTTPMiddleware


class Squall:
    _default_error_response = PlainTextResponse(
        "Internal Server Error", status_code=500
    )

    def __init__(
        self,
        *,
        debug: bool = False,
        routes: Optional[Optional[List[Union[APIRoute, WebSocketRoute]]]] = None,
        title: str = "Squall",
        description: str = "",
        version: str = "0.1.0",
        openapi_url: Optional[str] = "/openapi.json",
        openapi_tags: Optional[List[Dict[str, Any]]] = None,
        servers: Optional[List[Dict[str, Union[str, Any]]]] = None,
        default_response_class: Type[Response] = Default(JSONResponse),
        docs_url: Optional[str] = "/docs",
        redoc_url: Optional[str] = "/redoc",
        swagger_ui_oauth2_redirect_url: Optional[str] = "/docs/oauth2-redirect",
        swagger_ui_init_oauth: Optional[Dict[str, Any]] = None,
        middleware: Optional[Sequence[Middleware]] = None,
        exception_handlers: Optional[Dict[Union[int, Type[Exception]], AnyFunc]] = None,
        on_startup: Optional[Sequence[Callable[[], Any]]] = None,
        on_shutdown: Optional[Sequence[Callable[[], Any]]] = None,
        terms_of_service: Optional[str] = None,
        contact: Optional[Dict[str, Union[str, Any]]] = None,
        license_info: Optional[Dict[str, Union[str, Any]]] = None,
        openapi_prefix: str = "",
        root_path: str = "",
        root_path_in_servers: bool = True,
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        include_in_schema: bool = True,
        compression: Optional[Compression] = None,
        **extra: Any,
    ) -> None:
        self.debug: bool = debug
        self.state: State = State()

        self.router: router.RootRouter = router.RootRouter(
            routes=routes,
            default_response_class=default_response_class,
            deprecated=deprecated,
            include_in_schema=include_in_schema,
            responses=responses,
        )
        # Router methods linking for better user experience like having
        # @app.get(...) instead of @app.get(...)
        self.get = self.router.get
        self.put = self.router.put
        self.post = self.router.post
        self.delete = self.router.delete
        self.options = self.router.options
        self.head = self.router.head
        self.patch = self.router.patch
        self.trace = self.router.trace
        self.include_router = self.router.include_router
        self.routes = self.router.routes
        self.add_route = self.router.add_route
        self.add_api = self.router.add_api
        self.add_api_route = self.router.add_api_route
        self.websocket = self.router.websocket

        self.exception_handlers: Dict[Union[int, Type[Exception]], AnyFunc] = (
            {} if exception_handlers is None else dict(exception_handlers)
        )
        self.exception_handlers.setdefault(HTTPException, http_exception_handler)
        self.exception_handlers.setdefault(
            RequestPayloadValidationError, request_payload_validation_exception_handler
        )
        self.exception_handlers.setdefault(
            RequestHeadValidationError, request_head_validation_exception_handler
        )

        self.user_middleware: List[Middleware] = (
            [] if middleware is None else list(middleware)
        )
        self.middleware_stack: ASGIApp = self._build_middleware_stack()

        self.title = title
        self.description = description
        self.version = version
        self.terms_of_service = terms_of_service
        self.contact = contact
        self.license_info = license_info
        self.servers = servers or []
        self.openapi_url = openapi_url
        self.openapi_tags = openapi_tags
        # TODO: remove when discarding the openapi_prefix parameter
        if openapi_prefix:
            logger.warning(
                '"openapi_prefix" has been deprecated in favor of "root_path", which '
                "follows more closely the ASGI standard, is simpler, and more "
                "automatic. Check the docs at "
                "https://squall.mtag.dev/advanced/sub-applications/"
            )
        self.root_path = root_path or openapi_prefix
        self.root_path_in_servers = root_path_in_servers
        self.docs_url = docs_url
        self.redoc_url = redoc_url
        self.swagger_ui_oauth2_redirect_url = swagger_ui_oauth2_redirect_url
        self.swagger_ui_init_oauth = swagger_ui_init_oauth
        self.extra = extra
        self.openapi_version = "3.0.2"

        if self.openapi_url:
            assert self.title, "A title must be provided for OpenAPI, e.g.: 'My API'"
            assert self.version, "A version must be provided for OpenAPI, e.g.: '2.1.0'"
        self.openapi_schema: Optional[Dict[str, Any]] = None
        self.on_startup = [] if on_startup is None else list(on_startup)
        self.on_shutdown = [] if on_shutdown is None else list(on_shutdown)
        self.lifespan_ctx = LifespanContext(self.on_startup, self.on_shutdown)
        self.compression = compression

        self._setup()

    def add_event_handler(self, event: str, handler: AnyFunc) -> None:
        """Registrates event hooks.
        There are two evens availiable: startup, shutdown

        Examples:
            >>> app = Squall()
            >>>
            >>> def some_func_to_call_on_startup():
            >>>     pass
            >>>
            >>> async def some_func_to_call_on_shutdown():
            >>>     pass
            >>>
            >>> app.add_event_handler("startup", some_func_to_call_on_startup)
            >>> app.add_event_handler("shutdown", some_func_to_call_on_shutdown)
        """
        if event == "startup":
            self.on_startup.append(handler)
        elif event == "shutdown":
            self.on_shutdown.append(handler)

    def on_event(self, event_type: str) -> typing.Callable[..., Any]:
        """Decorator for event hook registration

        Examples:
            >>> app = Squall()
            >>>
            >>> @app.on_event("startup")
            >>> def some_func_to_call_on_startup():
            >>>     pass
            >>>
            >>> @app.on_event("shutdown")
            >>> async def some_func_to_call_on_shutdown():
            >>>     pass
        """

        def decorator(func: AnyFunc) -> AnyFunc:
            self.add_event_handler(event_type, func)
            return func

        return decorator

    def add_exception_handler(
        self,
        exc_class_or_status_code: typing.Union[int, typing.Type[Exception]],
        handler: typing.Callable[..., Any],
    ) -> None:
        """Adds exception handler.

        Examples:
            >>> app = Squall()
            >>>
            >>> class BackEndException(Exception): pass
            >>>
            >>> async def backend_exception_handler():
            >>>     return JSONResponse(
            >>>         status_code=500,
            >>>         content={"message": "BackEnd error"}
            >>>     )
            >>>
            >>> app.add_exception_handler(BackEndException, backend_exception_handler)
        """
        self.exception_handlers[exc_class_or_status_code] = handler

    def exception_handler(
        self, exc_class_or_status_code: typing.Union[int, typing.Type[Exception]]
    ) -> typing.Callable[..., Any]:
        """Decorator for exception handler registration.

        Examples:
            >>> app = Squall()
            >>>
            >>> class BackEndException(Exception): pass
            >>>
            >>> @app.exception_handler(BackEndException)
            >>> async def handle_backend_exception():
            >>>     return JSONResponse(
            >>>         status_code=500,
            >>>         content={"message": "BackEnd error"}
            >>>     )
            >>>
        """

        def decorator(func: typing.Callable[..., Any]) -> typing.Callable[..., Any]:
            self.add_exception_handler(exc_class_or_status_code, func)
            return func

        return decorator

    def add_middleware(self, middleware_class: type, **options: typing.Any) -> None:
        """Adds middleware.

        Examples:
            >>> from squall.middleware.httpsredirect import HTTPSRedirectMiddleware
            >>>
            >>> app = Squall()
            >>>
            >>> app.add_middleware(HTTPSRedirectMiddleware)
        """
        self.user_middleware.insert(0, Middleware(middleware_class, **options))
        self.middleware_stack = self._build_middleware_stack()

    def middleware(self, middleware_type: str) -> typing.Callable[..., Any]:
        """Decorator for middleware registration.

        Examples:
            >>> app = Squall()
            >>>
            >>> @app.middleware("http")
            >>> async def add_process_time_header(request: Request, call_next):
            >>>     start_time = time.time()
            >>>     response = await call_next(request)
            >>>     process_time = time.time() - start_time
            >>>     response.headers["X-Process-Time"] = str(process_time)
            >>>     return response
        """
        assert (
            middleware_type == "http"
        ), 'Currently only middleware("http") is supported.'

        def decorator(func: ASGIApp) -> ASGIApp:
            self.add_middleware(BaseHTTPMiddleware, dispatch=func)
            return func

        return decorator

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """ASGI calls entrypoint."""
        if self.root_path:
            scope["root_path"] = self.root_path
        scope["app"] = self

        if scope["type"] == "lifespan":
            await lifespan(self.lifespan_ctx, scope, receive, send)
            return

        try:
            await self.middleware_stack(scope, receive, send)
        except Exception as exc:
            handler = None

            if isinstance(exc, HTTPException):
                handler = self.exception_handlers.get(exc.status_code)

            if handler is None:
                handler = self._lookup_exception_handler(exc)

            request = Request(scope)
            if handler:
                if iscoroutinefunction(handler):
                    response = await handler(request, exc)
                else:
                    response = await run_in_threadpool(handler, request, exc)
            elif self.debug:
                # In debug mode, return traceback responses.
                response = get_default_debug_response(request, exc)
            else:
                # Use our default 500 error handler.
                response = self._default_error_response

            await response(scope, receive, send)

            if not handler:
                raise exc

    def _lookup_exception_handler(
        self, exc: Exception
    ) -> typing.Optional[typing.Callable[..., Any]]:
        exception_class = type(exc)
        try:
            return self.exception_handlers[exception_class]
        except KeyError:
            for cls in type(exc).__mro__:
                if cls in self.exception_handlers:
                    self.exception_handlers[exception_class] = self.exception_handlers[
                        cls
                    ]
                    return self.exception_handlers[cls]
        return None

    def _build_middleware_stack(self) -> ASGIApp:
        """Build stack for middlewares pipelining"""
        app = self.router
        for cls, options in reversed(self.user_middleware):
            app = cls(app=app, **options)
        return app

    def openapi(self) -> Dict[str, Any]:
        if not self.openapi_schema:
            self.openapi_schema = get_openapi(
                title=self.title,
                version=self.version,
                openapi_version=self.openapi_version,
                description=self.description,
                terms_of_service=self.terms_of_service,
                contact=self.contact,
                license_info=self.license_info,
                routes=self.routes,
                tags=self.openapi_tags,
                servers=self.servers,
            )
        return self.openapi_schema

    def _setup(self) -> None:
        """Setups OpenAPI functionality"""
        if self.openapi_url:
            urls = (server_data.get("url") for server_data in self.servers)
            server_urls = {url for url in urls if url}

            async def openapi(req: Request) -> JSONResponse:
                root_path = req.scope.get("root_path", "").rstrip("/")
                if root_path not in server_urls:
                    if root_path and self.root_path_in_servers:
                        self.servers.insert(0, {"url": root_path})
                        server_urls.add(root_path)
                return JSONResponse(self.openapi())

            self.router.add_api_route(
                self.openapi_url, openapi, include_in_schema=False
            )
        if self.openapi_url and self.docs_url:

            async def swagger_ui_html(req: Request) -> HTMLResponse:
                root_path = req.scope.get("root_path", "").rstrip("/")
                openapi_url = root_path + self.openapi_url
                oauth2_redirect_url = self.swagger_ui_oauth2_redirect_url
                if oauth2_redirect_url:
                    oauth2_redirect_url = root_path + oauth2_redirect_url
                return get_swagger_ui_html(
                    openapi_url=openapi_url,
                    title=self.title + " - Swagger UI",
                    oauth2_redirect_url=oauth2_redirect_url,
                    init_oauth=self.swagger_ui_init_oauth,
                )

            self.router.add_api_route(
                self.docs_url, swagger_ui_html, include_in_schema=False
            )

            if self.swagger_ui_oauth2_redirect_url:

                async def swagger_ui_redirect(req: Request) -> HTMLResponse:
                    return get_swagger_ui_oauth2_redirect_html()

                self.router.add_api_route(
                    self.swagger_ui_oauth2_redirect_url,
                    swagger_ui_redirect,
                    include_in_schema=False,
                )
        if self.openapi_url and self.redoc_url:

            async def redoc_html(req: Request) -> HTMLResponse:
                root_path = req.scope.get("root_path", "").rstrip("/")
                openapi_url = root_path + self.openapi_url
                return get_redoc_html(
                    openapi_url=openapi_url, title=self.title + " - ReDoc"
                )

            self.router.add_api_route(
                self.redoc_url, redoc_html, include_in_schema=False
            )



import pytest
from squall.testclient import TestClient

from .main import app

client = TestClient(app)

openapi_schema = {
    "openapi": "3.0.2",
    "info": {"title": "Squall", "version": "0.1.0"},
    "paths": {
        "/api_route": {
            "get": {
                "summary": "Non Operation",
                "operationId": "non_operation_api_route_get",
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    }
                },
            }
        },
        "/non_decorated_route": {
            "get": {
                "summary": "Non Decorated Route",
                "operationId": "non_decorated_route_non_decorated_route_get",
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    }
                },
            }
        },
        "/text": {
            "get": {
                "summary": "Get Text",
                "operationId": "get_text_text_get",
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    }
                },
            }
        },
        "/path/{item_id}": {
            "get": {
                "summary": "Get Id",
                "operationId": "get_id_path__item_id__get",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"type": "string"},
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/path/str/{item_id}": {
            "get": {
                "summary": "Get Str Id",
                "operationId": "get_str_id_path_str__item_id__get",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"type": "string"},
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/path/int/{item_id}": {
            "get": {
                "summary": "Get Int Id",
                "operationId": "get_int_id_path_int__item_id__get",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"type": "integer"},
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/path/float/{item_id}": {
            "get": {
                "summary": "Get Float Id",
                "operationId": "get_float_id_path_float__item_id__get",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"type": "number"},
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/path/param/{item_id}": {
            "get": {
                "summary": "Get Path Param Id",
                "operationId": "get_path_param_id_path_param__item_id__get",
                "parameters": [
                    {
                        "required": False,
                        "schema": {"type": "string"},
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/path/param-required/{item_id}": {
            "get": {
                "summary": "Get Path Param Required Id",
                "operationId": "get_path_param_required_id_path_param_required__item_id__get",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"type": "string"},
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/path/param-minlength/{item_id}": {
            "get": {
                "summary": "Get Path Param Min Len",
                "operationId": "get_path_param_min_len_path_param_minlength__item_id__get",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"type": "string"},
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/path/param-maxlength/{item_id}": {
            "get": {
                "summary": "Get Path Param Max Len",
                "operationId": "get_path_param_max_len_path_param_maxlength__item_id__get",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"type": "string"},
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/path/param-min_maxlength/{item_id}": {
            "get": {
                "summary": "Get Path Param Min Max Len",
                "operationId": "get_path_param_min_max_len_path_param_min_maxlength__item_id__get",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"type": "string"},
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/path/param-gt/{item_id}": {
            "get": {
                "summary": "Get Path Param Gt",
                "operationId": "get_path_param_gt_path_param_gt__item_id__get",
                "parameters": [
                    {
                        "required": True,
                        "schema": {
                            "type": "number",
                            "minimum": 3,
                            "exclusiveMinimum": True,
                        },
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/path/param-gt0/{item_id}": {
            "get": {
                "summary": "Get Path Param Gt0",
                "operationId": "get_path_param_gt0_path_param_gt0__item_id__get",
                "parameters": [
                    {
                        "required": True,
                        "schema": {
                            "type": "number",
                            "minimum": 0,
                            "exclusiveMinimum": True,
                        },
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/path/param-ge/{item_id}": {
            "get": {
                "summary": "Get Path Param Ge",
                "operationId": "get_path_param_ge_path_param_ge__item_id__get",
                "parameters": [
                    {
                        "required": True,
                        "schema": {
                            "type": "number",
                            "minimum": 3,
                            "exclusiveMinimum": False,
                        },
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/path/param-lt/{item_id}": {
            "get": {
                "summary": "Get Path Param Lt",
                "operationId": "get_path_param_lt_path_param_lt__item_id__get",
                "parameters": [
                    {
                        "required": True,
                        "schema": {
                            "type": "number",
                            "maximum": 3,
                            "exclusiveMaximum": True,
                        },
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/path/param-lt0/{item_id}": {
            "get": {
                "summary": "Get Path Param Lt0",
                "operationId": "get_path_param_lt0_path_param_lt0__item_id__get",
                "parameters": [
                    {
                        "required": True,
                        "schema": {
                            "type": "number",
                            "maximum": 0,
                            "exclusiveMaximum": True,
                        },
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/path/param-le/{item_id}": {
            "get": {
                "summary": "Get Path Param Le",
                "operationId": "get_path_param_le_path_param_le__item_id__get",
                "parameters": [
                    {
                        "required": True,
                        "schema": {
                            "type": "number",
                            "maximum": 3,
                            "exclusiveMaximum": False,
                        },
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/path/param-lt-gt/{item_id}": {
            "get": {
                "summary": "Get Path Param Lt Gt",
                "operationId": "get_path_param_lt_gt_path_param_lt_gt__item_id__get",
                "parameters": [
                    {
                        "required": True,
                        "schema": {
                            "type": "number",
                            "minimum": 1,
                            "exclusiveMinimum": True,
                            "maximum": 3,
                            "exclusiveMaximum": True,
                        },
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/path/param-le-ge/{item_id}": {
            "get": {
                "summary": "Get Path Param Le Ge",
                "operationId": "get_path_param_le_ge_path_param_le_ge__item_id__get",
                "parameters": [
                    {
                        "required": True,
                        "schema": {
                            "type": "number",
                            "minimum": 1,
                            "exclusiveMinimum": False,
                            "maximum": 3,
                            "exclusiveMaximum": False,
                        },
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/path/param-lt-int/{item_id}": {
            "get": {
                "summary": "Get Path Param Lt Int",
                "operationId": "get_path_param_lt_int_path_param_lt_int__item_id__get",
                "parameters": [
                    {
                        "required": True,
                        "schema": {
                            "type": "integer",
                            "maximum": 3,
                            "exclusiveMaximum": True,
                        },
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/path/param-gt-int/{item_id}": {
            "get": {
                "summary": "Get Path Param Gt Int",
                "operationId": "get_path_param_gt_int_path_param_gt_int__item_id__get",
                "parameters": [
                    {
                        "required": True,
                        "schema": {
                            "type": "integer",
                            "minimum": 3,
                            "exclusiveMinimum": True,
                        },
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/path/param-le-int/{item_id}": {
            "get": {
                "summary": "Get Path Param Le Int",
                "operationId": "get_path_param_le_int_path_param_le_int__item_id__get",
                "parameters": [
                    {
                        "required": True,
                        "schema": {
                            "type": "integer",
                            "maximum": 3,
                            "exclusiveMaximum": False,
                        },
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/path/param-ge-int/{item_id}": {
            "get": {
                "summary": "Get Path Param Ge Int",
                "operationId": "get_path_param_ge_int_path_param_ge_int__item_id__get",
                "parameters": [
                    {
                        "required": True,
                        "schema": {
                            "type": "integer",
                            "minimum": 3,
                            "exclusiveMinimum": False,
                        },
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/path/param-lt-gt-int/{item_id}": {
            "get": {
                "summary": "Get Path Param Lt Gt Int",
                "operationId": "get_path_param_lt_gt_int_path_param_lt_gt_int__item_id__get",
                "parameters": [
                    {
                        "required": True,
                        "schema": {
                            "type": "integer",
                            "minimum": 1,
                            "exclusiveMinimum": True,
                            "maximum": 3,
                            "exclusiveMaximum": True,
                        },
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/path/param-le-ge-int/{item_id}": {
            "get": {
                "summary": "Get Path Param Le Ge Int",
                "operationId": "get_path_param_le_ge_int_path_param_le_ge_int__item_id__get",
                "parameters": [
                    {
                        "required": True,
                        "schema": {
                            "type": "integer",
                            "minimum": 1,
                            "exclusiveMinimum": False,
                            "maximum": 3,
                            "exclusiveMaximum": False,
                        },
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/query": {
            "get": {
                "summary": "Get Query",
                "operationId": "get_query_query_get",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"type": "string"},
                        "name": "query",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/query/optional": {
            "get": {
                "summary": "Get Query Optional",
                "operationId": "get_query_optional_query_optional_get",
                "parameters": [
                    {
                        "required": False,
                        "schema": {"type": "string"},
                        "name": "query",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/query/int": {
            "get": {
                "summary": "Get Query Type",
                "operationId": "get_query_type_query_int_get",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"type": "integer"},
                        "name": "query",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/query/int/optional": {
            "get": {
                "summary": "Get Query Type Optional",
                "operationId": "get_query_type_optional_query_int_optional_get",
                "parameters": [
                    {
                        "required": False,
                        "schema": {"type": "integer"},
                        "name": "query",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/query/int/default": {
            "get": {
                "summary": "Get Query Type Int Default",
                "operationId": "get_query_type_int_default_query_int_default_get",
                "parameters": [
                    {
                        "required": False,
                        "schema": {"type": "integer"},
                        "name": "query",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/query/param": {
            "get": {
                "summary": "Get Query Param",
                "operationId": "get_query_param_query_param_get",
                "parameters": [
                    {
                        "required": False,
                        "schema": {"type": "string"},
                        "name": "query",
                        "in": "query",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/query/param-required": {
            "get": {
                "summary": "Get Query Param Required",
                "operationId": "get_query_param_required_query_param_required_get",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"type": "string"},
                        "name": "query",
                        "in": "query",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/query/param-required/int": {
            "get": {
                "summary": "Get Query Param Required Type",
                "operationId": "get_query_param_required_type_query_param_required_int_get",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"type": "integer"},
                        "name": "query",
                        "in": "query",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/enum-status-code": {
            "get": {
                "summary": "Get Enum Status Code",
                "operationId": "get_enum_status_code_enum_status_code_get",
                "responses": {
                    "201": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    }
                },
            }
        },
    },
}


@pytest.mark.parametrize(
    "path,expected_status,expected_response",
    [
        ("/api_route", 200, {"message": "Hello World"}),
        ("/non_decorated_route", 200, {"message": "Hello World"}),
        ("/nonexistent", 404, {"detail": "Not Found"}),
        ("/openapi.json", 200, openapi_schema),
    ],
)
def test_get_path(path, expected_status, expected_response):
    response = client.get(path)
    assert response.status_code == expected_status
    assert response.json() == expected_response


def test_swagger_ui():
    response = client.get("/docs")
    assert response.status_code == 200, response.text
    assert response.headers["content-type"] == "text/html; charset=utf-8"
    assert "swagger-ui-dist" in response.text
    assert (
        "oauth2RedirectUrl: window.location.origin + '/docs/oauth2-redirect'"
        in response.text
    )


def test_swagger_ui_oauth2_redirect():
    response = client.get("/docs/oauth2-redirect")
    assert response.status_code == 200, response.text
    assert response.headers["content-type"] == "text/html; charset=utf-8"
    assert "window.opener.swaggerUIRedirectOauth2" in response.text


def test_redoc():
    response = client.get("/redoc")
    assert response.status_code == 200, response.text
    assert response.headers["content-type"] == "text/html; charset=utf-8"
    assert "redoc@next" in response.text


def test_enum_status_code_response():
    response = client.get("/enum-status-code")
    assert response.status_code == 201, response.text
    assert response.json() == "foo bar"



import pytest
from squall import Router, Squall
from squall.testclient import TestClient

app = Squall()

router = Router(prefix="/prefix")


@router.get("")
def get_empty():
    return ["OK"]


app.include_router(router)


client = TestClient(app)


def test_use_empty():
    with client:
        response = client.get("/prefix")
        assert response.status_code == 200, response.text
        assert response.json() == ["OK"]

        response = client.get("/prefix/")
        assert response.status_code == 200, response.text
        assert response.json() == ["OK"]


@pytest.mark.skip(reason="Should we raise an exception in such cases?")
def test_include_empty():
    # if both include and router.path are empty - it should raise exception
    with pytest.raises(Exception):
        app.include_router(router)



from squall import Router, Squall, WebSocket
from squall.testclient import TestClient

router = Router()
prefix_router = Router(prefix="/prefix")
app = Squall()


@app.websocket("/")
async def index(websocket: WebSocket):
    await websocket.accept()
    await websocket.send_text("Hello, world!")
    await websocket.close()


@app.websocket("/some/{param}")
async def index_with_param(websocket: WebSocket, param: str):
    await websocket.accept()
    await websocket.send_text(f"Param: {param}")
    await websocket.close()


@router.websocket("/router")
async def routerindex(websocket: WebSocket):
    await websocket.accept()
    await websocket.send_text("Hello, router!")
    await websocket.close()


@prefix_router.websocket("/")
async def routerprefixindex(websocket: WebSocket):
    await websocket.accept()
    await websocket.send_text("Hello, router with prefix!")
    await websocket.close()


@router.websocket("/router2")
async def routerindex2(websocket: WebSocket):
    await websocket.accept()
    await websocket.send_text("Hello, router!")
    await websocket.close()


app.include_router(router)
app.include_router(prefix_router)


def test_app():
    client = TestClient(app)
    with client.websocket_connect("/") as websocket:
        data = websocket.receive_text()
        assert data == "Hello, world!"


def test_router():
    client = TestClient(app)
    with client.websocket_connect("/router") as websocket:
        data = websocket.receive_text()
        assert data == "Hello, router!"


def test_prefix_router():
    client = TestClient(app)
    with client.websocket_connect("/prefix/") as websocket:
        data = websocket.receive_text()
        assert data == "Hello, router with prefix!"


def test_router2():
    client = TestClient(app)
    with client.websocket_connect("/router2") as websocket:
        data = websocket.receive_text()
        assert data == "Hello, router!"


def test_get_path_param():
    client = TestClient(app)
    with client.websocket_connect("/some/test") as websocket:
        data = websocket.receive_text()
        assert data == "Param: test"



import inspect

from squall.openapi.docs import get_redoc_html, get_swagger_ui_html


def test_strings_in_generated_swagger():
    sig = inspect.signature(get_swagger_ui_html)
    swagger_js_url = sig.parameters.get("swagger_js_url").default  # type: ignore
    swagger_css_url = sig.parameters.get("swagger_css_url").default  # type: ignore
    swagger_favicon_url = sig.parameters.get("swagger_favicon_url").default  # type: ignore
    html = get_swagger_ui_html(openapi_url="/docs", title="title")
    body_content = html.body.decode()
    assert swagger_js_url in body_content
    assert swagger_css_url in body_content
    assert swagger_favicon_url in body_content


def test_strings_in_custom_swagger():
    swagger_js_url = "swagger_fake_file.js"
    swagger_css_url = "swagger_fake_file.css"
    swagger_favicon_url = "swagger_fake_file.png"
    html = get_swagger_ui_html(
        openapi_url="/docs",
        title="title",
        swagger_js_url=swagger_js_url,
        swagger_css_url=swagger_css_url,
        swagger_favicon_url=swagger_favicon_url,
    )
    body_content = html.body.decode()
    assert swagger_js_url in body_content
    assert swagger_css_url in body_content
    assert swagger_favicon_url in body_content


def test_strings_in_generated_redoc():
    sig = inspect.signature(get_redoc_html)
    redoc_js_url = sig.parameters.get("redoc_js_url").default  # type: ignore
    redoc_favicon_url = sig.parameters.get("redoc_favicon_url").default  # type: ignore
    html = get_redoc_html(openapi_url="/docs", title="title")
    body_content = html.body.decode()
    assert redoc_js_url in body_content
    assert redoc_favicon_url in body_content


def test_strings_in_custom_redoc():
    redoc_js_url = "fake_redoc_file.js"
    redoc_favicon_url = "fake_redoc_file.png"
    html = get_redoc_html(
        openapi_url="/docs",
        title="title",
        redoc_js_url=redoc_js_url,
        redoc_favicon_url=redoc_favicon_url,
    )
    body_content = html.body.decode()
    assert redoc_js_url in body_content
    assert redoc_favicon_url in body_content


def test_google_fonts_in_generated_redoc():
    body_with_google_fonts = get_redoc_html(
        openapi_url="/docs", title="title"
    ).body.decode()
    assert "fonts.googleapis.com" in body_with_google_fonts
    body_without_google_fonts = get_redoc_html(
        openapi_url="/docs", title="title", with_google_fonts=False
    ).body.decode()
    assert "fonts.googleapis.com" not in body_without_google_fonts



import pytest
from squall import Header, Squall
from squall.responses import (
    FileResponse,
    HTMLResponse,
    JSONResponse,
    PlainTextResponse,
    RedirectResponse,
    StreamingResponse,
)
from squall.testclient import TestClient

app = Squall()

html_response = """
    <html>
        <head>
            <title>HTML</title>
        </head>
        <body>
            <h1>HTML!</h1>
        </body>
    </html>
"""


def fake_video_streamer():
    for _ in range(3):
        yield b"mandarinki"


@app.get("/json_response")
def get_json_response():
    return JSONResponse(content={"msg": "mandarinki"})


@app.get("/json_response_class", response_class=JSONResponse)
def get_json_response_class():
    return {"msg": "mandarinki"}


@app.get("/html_response")
def get_html_response():
    return HTMLResponse(content=html_response)


@app.get("/html_response_class", response_class=HTMLResponse)
def get_html_response_class():
    return html_response


@app.get("/plaintext_response")
def get_plaintext_response():
    return PlainTextResponse(content="mandarinki")


@app.get("/plaintext_response_class", response_class=PlainTextResponse)
def get_plaintext_response_class():
    return "mandarinki"


@app.get("/streaming_response")
def get_streaming_response():
    return StreamingResponse(content=fake_video_streamer())


@app.get("/streaming_response_class", response_class=StreamingResponse)
def get_streaming_response_class():
    return fake_video_streamer()


@app.get("/file_response")
def get_file_response(response_file: str = Header()):
    return FileResponse(response_file)


@app.get("/file_response_class", response_class=FileResponse)
def get_file_response_class(response_file: str = Header()):
    return response_file


@app.get("/redirect_response")
def get_redirect_response():
    return RedirectResponse("https://mandarinki.com")


@app.get("/redirect_response_class", response_class=RedirectResponse)
def get_redirect_response_class():
    return "https://mandarinki.com"


client = TestClient(app)


@pytest.mark.parametrize("path", ["/json_response", "/json_response_class"])
def test_json_response(path):
    with client:
        response = client.get(path)
    assert response.json() == {"msg": "mandarinki"}
    assert response.headers["content-type"] == "application/json"


@pytest.mark.parametrize("path", ["/html_response", "/html_response_class"])
def test_html_response(path):
    with client:
        response = client.get(path)
    assert response.content.decode() == html_response
    assert "text/html" in response.headers["content-type"]


@pytest.mark.parametrize("path", ["/plaintext_response", "/plaintext_response_class"])
def test_plaintext_response(path):
    with client:
        response = client.get(path)
    assert response.content.decode() == "mandarinki"
    assert "text/plain" in response.headers["content-type"]


@pytest.mark.parametrize("path", ["/streaming_response", "/streaming_response_class"])
def test_streaming_response(path):
    with client:
        response = client.get(path)
    assert response.content.decode() == "mandarinki" * 3


@pytest.mark.parametrize("path", ["/file_response", "/file_response_class"])
def test_file_response(path, response_file):
    with client:
        response = client.get(path, headers={"response_file": str(response_file)})
    assert response.content == b"salokovbasasaltison"
    assert "text/plain" in response.headers["content-type"]


@pytest.mark.parametrize("path", ["/redirect_response", "/redirect_response_class"])
def test_redirect_response(path):
    with client:
        response = client.get(path, allow_redirects=False)
    assert response.status_code == 307
    assert response.headers["location"] == "https://mandarinki.com"



from squall import Squall
from squall.testclient import TestClient

app = Squall()


@app.get("/", openapi_extra={"x-custom-extension": "value"})
def route_with_extras():
    return {}


client = TestClient(app)


openapi_schema = {
    "openapi": "3.0.2",
    "info": {"title": "Squall", "version": "0.1.0"},
    "paths": {
        "/": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                },
                "summary": "Route With Extras",
                "operationId": "route_with_extras__get",
                "x-custom-extension": "value",
            }
        },
    },
}


def test_openapi():
    response = client.get("/openapi.json")
    assert response.status_code == 200, response.text
    assert response.json() == openapi_schema


def test_get_route():
    response = client.get("/")
    assert response.status_code == 200, response.text
    assert response.json() == {}



# from squall import APIRouter, Request, Squall
# from squall.responses import JSONResponse
# from squall.testclient import TestClient
#
# app = Squall()
# router = APIRouter()
#
#
# @router.route("/items/")
# def read_items(request: Request):
#     return JSONResponse({"hello": "world"})
#
#
# app.include_router(router)
#
# client = TestClient(app)
#
#
# def test_sub_router():
#     response = client.get("/items/")
#     assert response.status_code == 200, response.text
#     assert response.json() == {"hello": "world"}



import pytest
from squall import File, Form, Squall, UploadFile
from squall.testclient import TestClient

app = Squall()


@app.post("/files/")
async def create_file(
    file: UploadFile = File(...), fileb: UploadFile = File(...), token: str = Form(...)
):
    file = await file.read()
    return {
        "file_size": len(file),
        "token": token,
        "fileb_content_type": fileb.content_type,
    }


client = TestClient(app)

openapi_schema = {
    "openapi": "3.0.2",
    "info": {"title": "FastAPI", "version": "0.1.0"},
    "paths": {
        "/files/": {
            "post": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "422": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Create File",
                "operationId": "create_file_files__post",
                "requestBody": {
                    "content": {
                        "multipart/form-data": {
                            "schema": {
                                "$ref": "#/components/schemas/Body_create_file_files__post"
                            }
                        }
                    },
                    "required": True,
                },
            }
        }
    },
    "components": {
        "schemas": {
            "Body_create_file_files__post": {
                "title": "Body_create_file_files__post",
                "required": ["file", "fileb", "token"],
                "type": "object",
                "properties": {
                    "file": {"title": "File", "type": "string", "format": "binary"},
                    "fileb": {"title": "Fileb", "type": "string", "format": "binary"},
                    "token": {"title": "Token", "type": "string"},
                },
            },
            "ValidationError": {
                "title": "ValidationError",
                "required": ["loc", "msg", "type"],
                "type": "object",
                "properties": {
                    "loc": {
                        "title": "Location",
                        "type": "array",
                        "items": {"type": "string"},
                    },
                    "msg": {"title": "Message", "type": "string"},
                    "type": {"title": "Error Type", "type": "string"},
                },
            },
            "HTTPValidationError": {
                "title": "HTTPValidationError",
                "type": "object",
                "properties": {
                    "detail": {
                        "title": "Detail",
                        "type": "array",
                        "items": {"$ref": "#/components/schemas/ValidationError"},
                    }
                },
            },
        }
    },
}


@pytest.mark.skip("Not implemented")
def test_openapi_schema():
    response = client.get("/openapi.json")
    assert response.status_code == 200, response.text
    assert response.json() == openapi_schema


file_required = {
    "details": [
        {"loc": ["form", "file"], "msg": "field required"},
        {"loc": ["form", "fileb"], "msg": "field required"},
    ]
}

token_required = {
    "details": [
        {"loc": ["form", "fileb"], "msg": "field required"},
        {"loc": ["form", "token"], "msg": "field required"},
    ]
}


file_and_token_required = {
    "details": [
        {"loc": ["form", "file"], "msg": "field required"},
        {"loc": ["form", "fileb"], "msg": "field required"},
        {"loc": ["form", "token"], "msg": "field required"},
    ]
}


def test_post_form_no_body():
    response = client.post("/files/")
    assert response.status_code == 422, response.text
    assert response.json() == file_and_token_required


def test_post_form_no_file():
    response = client.post("/files/", data={"token": "foo"})
    assert response.status_code == 422, response.text
    assert response.json() == file_required


def test_post_body_json():
    response = client.post("/files/", json={"file": "Foo", "token": "Bar"})
    assert response.status_code == 422, response.text
    assert response.json() == file_and_token_required


def test_post_file_no_token(tmp_path):
    path = tmp_path / "test.txt"
    path.write_bytes(b"<file content>")

    client = TestClient(app)
    with path.open("rb") as file:
        response = client.post("/files/", files={"file": file})
    assert response.status_code == 422, response.text
    assert response.json() == token_required


def test_post_files_and_token(tmp_path):
    patha = tmp_path / "test.txt"
    pathb = tmp_path / "testb.txt"
    patha.write_text("<file content>")
    pathb.write_text("<file b content>")

    client = TestClient(app)
    with patha.open("rb") as filea, pathb.open("rb") as fileb:
        response = client.post(
            "/files/",
            data={"token": "foo"},
            files={"file": filea, "fileb": ("testb.txt", fileb, "text/plain")},
        )
    assert response.status_code == 200, response.text
    assert response.json() == {
        "file_size": 14,
        "token": "foo",
        "fileb_content_type": "text/plain",
    }



from dataclasses import dataclass
from typing import Dict

from squall import Squall
from squall.testclient import TestClient

app = Squall()


@dataclass
class Items:
    items: Dict[str, int]


@app.post("/foo")
def foo(items: Items):
    return items.items


client = TestClient(app)


openapi_schema = {
    "openapi": "3.0.2",
    "info": {"title": "Squall", "version": "0.1.0"},
    "components": {
        "schemas": {
            "Items": {
                "type": "object",
                "properties": {
                    "items": {
                        "type": "object",
                        "additionalProperties": {"type": "integer"},
                    }
                },
                "required": ["items"],
                "additionalProperties": False,
            },
            "ValidationError": {
                "title": "ValidationError",
                "type": "object",
                "properties": {
                    "loc": {
                        "title": "Location",
                        "type": "array",
                        "items": {"type": "string"},
                    },
                    "msg": {"title": "Message", "type": "string"},
                    "type": {"title": "Error Type", "type": "string"},
                },
                "required": ["loc", "msg", "type"],
            },
            "HTTPValidationError": {
                "title": "HTTPValidationError",
                "type": "object",
                "properties": {
                    "detail": {
                        "title": "Detail",
                        "type": "array",
                        "items": {"$ref": "#/components/schemas/ValidationError"},
                    }
                },
            },
            "HTTPBadRequestError": {
                "title": "HTTPBadRequestError",
                "type": "object",
                "properties": {
                    "details": {
                        "title": "Detail",
                        "type": "array",
                        "items": {"$ref": "#/components/schemas/ValidationError"},
                    }
                },
            },
        }
    },
    "paths": {
        "/foo": {
            "post": {
                "summary": "Foo",
                "operationId": "foo_foo_post",
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "422": {
                        "description": "Request Body Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "requestBody": {
                    "required": True,
                    "content": {
                        "application/json": {
                            "schema": {"$ref": "#/components/schemas/Items"}
                        }
                    },
                },
            }
        }
    },
}


def test_additional_properties_schema():
    response = client.get("/openapi.json")
    assert response.status_code == 200, response.text
    assert response.json() == openapi_schema


def test_additional_properties_post():
    response = client.post("/foo", json={"items": {"foo": 1, "bar": 2}})
    assert response.status_code == 200, response.text
    assert response.json() == {"foo": 1, "bar": 2}



import pytest
from squall import Squall
from squall.testclient import TestClient

app = Squall()


@app.get("/a", responses={"hello": {"description": "Not a valid additional response"}})
async def a():
    pass  # pragma: no cover


openapi_schema = {
    "openapi": "3.0.2",
    "info": {"title": "Squall", "version": "0.1.0"},
    "paths": {
        "/a": {
            "get": {
                "responses": {
                    # this is how one would imagine the openapi schema to be
                    # but since the key is not valid, openapi.utils.get_openapi will raise ValueError
                    "hello": {"description": "Not a valid additional response"},
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                },
                "summary": "A",
                "operationId": "a_a_get",
            }
        }
    },
}

client = TestClient(app)


def test_openapi_schema():
    with pytest.raises(ValueError):
        client.get("/openapi.json")



# from typing import Any, List
#
# import pytest
# from squall.params import Body, Cookie, Depends, Header, Param, Path, Query
#
# test_data: List[Any] = ["teststr", None, ..., 1, []]
#
#
# def get_user():
#     return {}  # pragma: no cover
#
#
# @pytest.fixture(scope="function", params=test_data)
# def params(request):
#     return request.param
#
#
# def test_param_repr(params):
#     assert repr(Param(params)) == "Param(" + str(params) + ")"
#
#
# def test_path_repr(params):
#     assert repr(Path(params)) == "Path(Ellipsis)"
#
#
# def test_query_repr(params):
#     assert repr(Query(params)) == "Query(" + str(params) + ")"
#
#
# def test_header_repr(params):
#     assert repr(Header(params)) == "Header(" + str(params) + ")"
#
#
# def test_cookie_repr(params):
#     assert repr(Cookie(params)) == "Cookie(" + str(params) + ")"
#
#
# def test_body_repr(params):
#     assert repr(Body(params)) == "Body(" + str(params) + ")"
#
#
# def test_depends_repr():
#     assert repr(Depends()) == "Depends(NoneType)"
#     assert repr(Depends(get_user)) == "Depends(get_user)"
#     assert repr(Depends(use_cache=False)) == "Depends(NoneType, use_cache=False)"
#     assert (
#         repr(Depends(get_user, use_cache=False)) == "Depends(get_user, use_cache=False)"
#     )



from dataclasses import dataclass, field
from typing import List

from squall import Squall
from squall.testclient import TestClient

app = Squall()


@dataclass
class RecursiveItem:
    name: str = field()
    sub_items: List["RecursiveItem"] = field(default_factory=list)


@dataclass
class RecursiveSubitemInSubmodel:
    name: str = field()
    sub_items2: List["RecursiveItemViaSubmodel"] = field(default_factory=list)


@dataclass
class RecursiveItemViaSubmodel:
    name: str = field()
    sub_items1: List[RecursiveSubitemInSubmodel] = field(default_factory=list)


@app.get("/items/recursive", response_model=RecursiveItem)
def get_recursive():
    return {"name": "item", "sub_items": [{"name": "subitem", "sub_items": []}]}


@app.get("/items/recursive-submodel", response_model=RecursiveItemViaSubmodel)
def get_recursive_submodel():
    return {
        "name": "item",
        "sub_items1": [
            {
                "name": "subitem",
                "sub_items2": [
                    {
                        "name": "subsubitem",
                        "sub_items1": [{"name": "subsubsubitem", "sub_items2": []}],
                    }
                ],
            }
        ],
    }


client = TestClient(app)


def test_recursive():
    response = client.get("/items/recursive")
    assert response.status_code == 200, response.text
    assert response.json() == {
        "sub_items": [{"name": "subitem", "sub_items": []}],
        "name": "item",
    }

    response = client.get("/items/recursive-submodel")
    assert response.status_code == 200, response.text
    assert response.json() == {
        "name": "item",
        "sub_items1": [
            {
                "name": "subitem",
                "sub_items2": [
                    {
                        "name": "subsubitem",
                        "sub_items1": [{"name": "subsubsubitem", "sub_items2": []}],
                    }
                ],
            }
        ],
    }



import asyncio
import enum
import functools
import inspect
import re
from decimal import Decimal
from typing import (
    Any,
    Callable,
    Coroutine,
    Dict,
    List,
    Optional,
    Pattern,
    Set,
    Tuple,
    Type,
    Union,
)

from apischema import ValidationError, deserialization_method, serialization_method
from orjson import JSONDecodeError
from squall.bindings import RequestField, ResponseField
from squall.datastructures import Default, DefaultPlaceholder
from squall.exceptions import (
    HTTPException,
    RequestHeadValidationError,
    RequestPayloadValidationError,
    ResponsePayloadValidationError,
    WebSocketRequestValidationError,
)
from squall.requests import Request
from squall.responses import JSONResponse, Response
from squall.routing_.utils import (
    HeadParam,
    get_handler_body_params,
    get_handler_head_params,
    get_handler_request_fields,
)
from squall.types import ASGIApp
from squall.utils import generate_operation_id_for_path, get_callable_name
from squall.validators.head import Validator
from squall.websockets import WebSocket
from starlette.concurrency import run_in_threadpool
from starlette.convertors import CONVERTOR_TYPES, Convertor
from starlette.routing import websocket_session
from starlette.status import WS_1008_POLICY_VIOLATION
from starlette.types import Receive, Scope, Send


class BaseRoute:
    def __init__(
        self,
        path: str,
        endpoint: Callable[..., Any],
        *,
        methods: Optional[List[str]] = None,
    ) -> None:
        self.path = self._path_origin = path
        self.endpoint = endpoint
        self.methods = methods

    def matches(self, scope: Scope) -> Tuple[bool, Scope]:
        if match := self.path_regex.match(scope["path"]):
            return True, {"endpoint": self.endpoint, "path_params": match.groupdict()}
        return False, {}

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, BaseRoute)
            and self.path == other.path
            and self.endpoint == other.endpoint
            and self.methods == other.methods
        )

    def add_path_prefix(self, prefix: str) -> None:
        self.path = prefix + self.path
        self.path_regex, self.path_format, self.param_convertors = compile_path(
            self.path
        )

    def set_defaults(self) -> None:
        self.path = self._path_origin


class WebSocketRoute(BaseRoute):
    def __init__(
        self,
        path: str,
        endpoint: Callable[..., Any],
        *,
        name: Optional[str] = None,
        include_in_schema: bool = True,
    ) -> None:
        self.path = self._path_origin = path
        self.endpoint = endpoint
        self.head_params = get_handler_head_params(endpoint)
        self.head_validator = build_head_validator(self.head_params)
        self.name = get_callable_name(endpoint) if name is None else name
        self.include_in_schema = include_in_schema

        if inspect.isfunction(endpoint) or inspect.ismethod(endpoint):
            # Endpoint is function or method. Treat it as `func(websocket)`.
            self.app = websocket_session(
                get_websocket_handler(endpoint, head_validator=self.head_validator)
            )
        else:
            # Endpoint is a class. Treat it as ASGI.
            self.app = endpoint

        self.path_regex, self.path_format, self.param_convertors = compile_path(path)


class NoMatchFound(Exception):
    """
    Raised by `.url_for(name, **path_params)` and `.url_path_for(name, **path_params)`
    if no matching route exists.
    """


# Match parameters in URL paths, eg. '{param}', and '{param:int}'
PARAM_REGEX = re.compile("{([a-zA-Z_][a-zA-Z0-9_]*)(:[a-zA-Z_][a-zA-Z0-9_]*)?}")


def compile_path(
    path: str,
) -> Tuple[Pattern[str], str, Dict[str, Convertor]]:
    """
    Given a path string, like: "/{username:str}", return a three-tuple
    of (regex, format, {param_name:convertor}).
    regex:      "/(?P<username>[^/]+)"
    format:     "/{username}"
    convertors: {"username": StringConvertor()}
    """
    path_regex = "^"
    path_format = ""
    duplicated_params = set()

    idx = 0
    param_convertors = {}
    for match in PARAM_REGEX.finditer(path):
        param_name, convertor_type = match.groups("str")
        convertor_type = convertor_type.lstrip(":")
        assert (
            convertor_type in CONVERTOR_TYPES
        ), f"Unknown path convertor '{convertor_type}'"
        convertor = CONVERTOR_TYPES[convertor_type]

        path_regex += re.escape(path[idx : match.start()])
        path_regex += f"(?P<{param_name}>{convertor.regex})"

        path_format += path[idx : match.start()]
        path_format += "{%s}" % param_name

        if param_name in param_convertors:
            duplicated_params.add(param_name)

        param_convertors[param_name] = convertor

        idx = match.end()

    if duplicated_params:
        names = ", ".join(sorted(duplicated_params))
        ending = "s" if len(duplicated_params) > 1 else ""
        raise ValueError(f"Duplicated param name{ending} {names} at path {path}")

    path_regex += re.escape(path[idx:]) + "$"
    path_format += path[idx:]

    return re.compile(path_regex), path_format, param_convertors


def get_http_handler(
    endpoint: Callable[..., Any],
    head_validator: Optional[Callable[..., Any]] = None,
    body_fields: Optional[List[Any]] = None,
    status_code: Optional[int] = None,
    response_class: Union[Type[Response], DefaultPlaceholder] = Default(JSONResponse),
    request_field: Optional[RequestField] = None,
    request_deserializer: Optional[Callable[..., Any]] = None,
    response_deserializer: Optional[Callable[..., Any]] = None,
    response_serializer: Optional[Callable[..., Any]] = None,
) -> ASGIApp:
    is_coroutine = asyncio.iscoroutinefunction(endpoint)
    # is_body_form = body_field and isinstance(body_field.field_info, params.Form)
    if isinstance(response_class, DefaultPlaceholder):
        actual_response_class: Type[Response] = response_class.value
    else:
        actual_response_class = response_class

    request_model = request_model_param = None
    if request_field is not None:
        request_model_param = request_field.name
        request_model = request_field.model

    async def app(scope: Scope, receive: Receive, send: Send) -> None:
        request = Request(scope, receive=receive, send=send)

        # Head validation
        if head_validator is not None:
            kwargs, errors = head_validator(request)
            if errors:
                raise RequestHeadValidationError(errors)
        else:
            kwargs = {}

        # Body fields and request object
        form = None
        try:
            if request_model is not None and request_deserializer is not None:
                body = await request.json()
                kwargs[request_model_param] = request_deserializer(body)

            if body_fields:
                form_missed = []
                for field in body_fields:
                    kind = field["kind"]
                    if kind == "request":
                        kwargs[field["name"]] = request
                    elif kind == "body":
                        ct = request.headers.get("content-type")
                        if ct is not None and ct[-4:] == "json":
                            kwargs[field["name"]] = await request.json()
                        else:
                            kwargs[field["name"]] = await request.body()
                    elif kind == "form":
                        if form is None:
                            form = await request.form()
                        value = form.get(field["name"])
                        if value is None:
                            form_missed.append(
                                {
                                    "loc": ["form", field["name"]],
                                    "msg": "field required",
                                }
                            )
                        else:
                            kwargs[field["name"]] = form.get(field["name"])
                if form_missed:
                    raise ValidationError.from_errors(form_missed)  # type: ignore
        except JSONDecodeError as e:
            raise RequestPayloadValidationError([str(e)])
        except ValidationError as e:
            raise RequestPayloadValidationError(e.messages, e.children)
        except Exception as e:
            raise HTTPException(
                status_code=400, detail="There was an error parsing the body"
            ) from e

        if is_coroutine:
            raw_response = await endpoint(**kwargs)
        else:
            raw_response = await run_in_threadpool(endpoint, **kwargs)

        if isinstance(raw_response, Response):
            raw_response.request = request
            await raw_response(scope, receive, send)
            return

        response_args: Dict[str, Any] = {}

        # If status_code was set, use it, otherwise use the default from the
        # response class, in the case of redirect it's 307

        try:
            if response_deserializer is not None and response_serializer is not None:
                result = response_serializer(response_deserializer(raw_response))
            elif response_serializer is not None:
                result = response_serializer(raw_response)
            else:
                result = raw_response
        except ValidationError as e:
            raise ResponsePayloadValidationError(e.messages, e.children)
        except TypeError as e:
            raise ResponsePayloadValidationError([str(e)])

        if status_code is not None:
            response_args["status_code"] = status_code
        response = actual_response_class(result, **response_args)
        # Temporary solution in order to avoid header initialization
        response.request = request
        await response(scope, receive, send)

    return app


def get_websocket_handler(
    endpoint: Callable[..., Any],
    head_validator: Optional[Callable[..., Any]] = None,
) -> Callable[[WebSocket], Coroutine[Any, Any, Any]]:
    async def app(websocket: WebSocket) -> None:
        # Head validation
        if head_validator is not None:
            kwargs, errors = head_validator(websocket)
            if errors:
                await websocket.close(code=WS_1008_POLICY_VIOLATION)
                raise WebSocketRequestValidationError(errors)
        else:
            kwargs = {}
        await endpoint(websocket, **kwargs)

    return app


def build_head_validator(head_params: List[HeadParam]) -> Callable[..., Any]:
    v = Validator(
        args=["request"],
        convertors={
            "int": int,
            "float": float,
            "Decimal": Decimal,
            "str": lambda a: a.decode("utf-8") if type(a) == bytes else str(a),
            "bytes": lambda a: str(a).encode("utf-8") if type(a) != bytes else a,
        },
    )
    for param in head_params:
        v.add_rule(
            attribute=param.source,
            name=param.name,
            key=param.alias,
            check=param.validate,
            convert=param.convertor,
            as_list=param.is_array,
            default=param.default,
            **param.statements,
        )
    return v.build()


class Route(BaseRoute):
    def __init__(
        self,
        path: str,
        endpoint: Callable[..., Any],
        *,
        methods: Optional[List[str]] = None,
        name: Optional[str] = None,
        include_in_schema: bool = True,
    ) -> None:
        assert path.startswith("/"), "Routed paths must start with '/'"
        self.path = path
        self.endpoint = endpoint
        self.name = get_callable_name(endpoint) if name is None else name
        self.include_in_schema = include_in_schema

        endpoint_handler = endpoint
        while isinstance(endpoint_handler, functools.partial):
            endpoint_handler = endpoint_handler.func
        if inspect.isfunction(endpoint_handler) or inspect.ismethod(endpoint_handler):
            # Endpoint is function or method. Treat it as `func(request) -> response`.
            self.app = get_http_handler(endpoint=endpoint)
            if methods is None:
                methods = ["GET"]
        else:
            # Endpoint is a class. Treat it as ASGI.
            self.app = endpoint

        if methods is None:
            self.methods = None
        else:
            self.methods = [method.upper() for method in methods]
            if self.methods is not None and "GET" in self.methods:
                self.methods.append("HEAD")

        self.path_regex, self.path_format, self.param_convertors = compile_path(path)


class APIRoute(Route):
    def __init__(
        self,
        path: str,
        endpoint: Callable[..., Any],
        *,
        response_model: Optional[Type[Any]] = None,
        status_code: Optional[int] = None,
        tags: Optional[List[str]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_description: str = "Successful Response",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        name: Optional[str] = None,
        methods: Optional[Union[Set[str], List[str]]] = None,
        operation_id: Optional[str] = None,
        include_in_schema: bool = True,
        response_class: Union[Type[Response], DefaultPlaceholder] = Default(
            JSONResponse
        ),
        openapi_extra: Optional[Dict[str, Any]] = None,
    ) -> None:
        # normalise enums e.g. http.HTTPStatus
        if isinstance(status_code, enum.IntEnum):
            status_code = int(status_code)
        self.path = self._path_origin = path
        self.endpoint = endpoint
        self.head_params = get_handler_head_params(endpoint)
        self.head_validator = build_head_validator(self.head_params)

        self.body_fields = get_handler_body_params(endpoint)

        request_fields = get_handler_request_fields(endpoint)
        assert len(request_fields) < 2, "Only one request model allowed"
        self.request_field: Optional[RequestField] = None
        self.request_deserializer: Optional[Callable[..., Any]] = None
        if request_fields:
            self.request_field = request_fields[0]
            self.request_deserializer = deserialization_method(self.request_field.model)

        self.name = get_callable_name(endpoint) if name is None else name
        self.path_regex, self.path_format, self.param_convertors = compile_path(path)
        if methods is None:
            methods = ["GET"]
        self.methods = [method.upper() for method in methods]
        self.unique_id = generate_operation_id_for_path(
            name=self.name, path=self.path_format, method=list(methods)[0]
        )
        self.response_field: Optional[ResponseField] = None
        self.response_deserializer: Optional[Callable[..., Any]] = None
        self.response_serializer: Optional[Callable[..., Any]] = None

        endpoint_returns = inspect.signature(endpoint).return_annotation
        res_deserialize = True
        if response_model == endpoint_returns:
            res_deserialize = False
        # elif endpoint_returns != inspect._empty and response_model is None:
        #     response_model = endpoint_returns
        #     res_deserialize = False

        if response_model is not None:
            self.response_field = ResponseField(model=response_model)
            check_type_on_serialization = True
            if res_deserialize:
                self.response_deserializer = deserialization_method(
                    self.response_field.model
                )
                check_type_on_serialization = False
            self.response_serializer = serialization_method(
                self.response_field.model, check_type=check_type_on_serialization
            )

        self.status_code = status_code
        self.tags = tags or []
        self.summary = summary
        self.description = description or inspect.cleandoc(self.endpoint.__doc__ or "")
        self.description = self.description.split("\f")[0]
        self.response_description = response_description
        self.responses = responses or {}
        self.deprecated = deprecated
        self.operation_id = operation_id
        self.include_in_schema = include_in_schema
        self.response_class = response_class

        assert callable(endpoint), "An endpoint must be a callable"

        self.app = self.get_route_handler()
        self.openapi_extra = openapi_extra

    def get_route_handler(self) -> ASGIApp:
        return get_http_handler(
            endpoint=self.endpoint,
            status_code=self.status_code,
            response_class=self.response_class,
            request_field=self.request_field,
            request_deserializer=self.request_deserializer,
            response_deserializer=self.response_deserializer,
            response_serializer=self.response_serializer,
            head_validator=self.head_validator,
            body_fields=self.body_fields,
        )



from dataclasses import dataclass, field
from typing import Optional

from squall import Squall
from squall.responses import JSONResponse
from squall.testclient import TestClient

app = Squall()


@dataclass
class Item:
    name: str = field()
    price: Optional[float] = None


@app.add_api("/items/{item_id}", methods=["GET"])
def get_items(item_id: str):
    return {"item_id": item_id}


def get_not_decorated(item_id: str):
    return {"item_id": item_id}


app.add_api_route("/items-not-decorated/{item_id}", get_not_decorated)


@app.delete("/items/{item_id}")
def delete_item(item_id: str, item: Item):
    return {"item_id": item_id, "item": item}


@app.head("/items/{item_id}")
def head_item(item_id: str):
    return JSONResponse(headers={"x-squall-item-id": item_id})


@app.options("/items/{item_id}")
def options_item(item_id: str):
    return JSONResponse(headers={"x-squall-item-id": item_id})


@app.patch("/items/{item_id}")
def patch_item(item_id: str, item: Item):
    return {"item_id": item_id, "item": item}


@app.trace("/items/{item_id}")
def trace_item(item_id: str):
    return JSONResponse(media_type="message/http")


client = TestClient(app)

openapi_schema = {
    "openapi": "3.0.2",
    "info": {"title": "Squall", "version": "0.1.0"},
    "components": {
        "schemas": {
            "Item": {
                "type": "object",
                "properties": {
                    "name": {"type": "string"},
                    "price": {"type": ["number", "null"], "default": None},
                },
                "required": ["name"],
                "additionalProperties": False,
            },
            "ValidationError": {
                "title": "ValidationError",
                "type": "object",
                "properties": {
                    "loc": {
                        "title": "Location",
                        "type": "array",
                        "items": {"type": "string"},
                    },
                    "msg": {"title": "Message", "type": "string"},
                    "type": {"title": "Error Type", "type": "string"},
                },
                "required": ["loc", "msg", "type"],
            },
            "HTTPValidationError": {
                "title": "HTTPValidationError",
                "type": "object",
                "properties": {
                    "detail": {
                        "title": "Detail",
                        "type": "array",
                        "items": {"$ref": "#/components/schemas/ValidationError"},
                    }
                },
            },
            "HTTPBadRequestError": {
                "title": "HTTPBadRequestError",
                "type": "object",
                "properties": {
                    "details": {
                        "title": "Detail",
                        "type": "array",
                        "items": {"$ref": "#/components/schemas/ValidationError"},
                    }
                },
            },
        }
    },
    "paths": {
        "/items/{item_id}": {
            "get": {
                "summary": "Get Items",
                "operationId": "get_items_items__item_id__get",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"type": "string"},
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            },
            "delete": {
                "summary": "Delete Item",
                "operationId": "delete_item_items__item_id__delete",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"type": "string"},
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                    "422": {
                        "description": "Request Body Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "requestBody": {
                    "required": True,
                    "content": {
                        "application/json": {
                            "schema": {"$ref": "#/components/schemas/Item"}
                        }
                    },
                },
            },
            "head": {
                "summary": "Head Item",
                "operationId": "head_item_items__item_id__head",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"type": "string"},
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            },
            "options": {
                "summary": "Options Item",
                "operationId": "options_item_items__item_id__options",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"type": "string"},
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            },
            "patch": {
                "summary": "Patch Item",
                "operationId": "patch_item_items__item_id__patch",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"type": "string"},
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                    "422": {
                        "description": "Request Body Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "requestBody": {
                    "required": True,
                    "content": {
                        "application/json": {
                            "schema": {"$ref": "#/components/schemas/Item"}
                        }
                    },
                },
            },
            "trace": {
                "summary": "Trace Item",
                "operationId": "trace_item_items__item_id__trace",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"type": "string"},
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            },
        },
        "/items-not-decorated/{item_id}": {
            "get": {
                "summary": "Get Not Decorated",
                "operationId": "get_not_decorated_items_not_decorated__item_id__get",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"type": "string"},
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
    },
}


def test_openapi_schema():
    response = client.get("/openapi.json")
    assert response.status_code == 200, response.text
    assert response.json() == openapi_schema


def test_get_api_route():
    response = client.get("/items/foo")
    assert response.status_code == 200, response.text
    assert response.json() == {"item_id": "foo"}


def test_get_api_route_not_decorated():
    response = client.get("/items-not-decorated/foo")
    assert response.status_code == 200, response.text
    assert response.json() == {"item_id": "foo"}


def test_delete():
    response = client.delete("/items/foo", json={"name": "Foo"})
    assert response.status_code == 200, response.text
    assert response.json() == {"item_id": "foo", "item": {"name": "Foo", "price": None}}


def test_head():
    response = client.head("/items/foo")
    assert response.status_code == 200, response.text
    assert response.headers["x-squall-item-id"] == "foo"


def test_options():
    response = client.options("/items/foo")
    assert response.status_code == 200, response.text
    assert response.headers["x-squall-item-id"] == "foo"


def test_patch():
    response = client.patch("/items/foo", json={"name": "Foo"})
    assert response.status_code == 200, response.text
    assert response.json() == {"item_id": "foo", "item": {"name": "Foo", "price": None}}


def test_trace():
    response = client.request("trace", "/items/foo")
    assert response.status_code == 200, response.text
    assert response.headers["content-type"] == "message/http"
